import gc
import glob
import hashlib
import json
import os
import re
import subprocess
import xml.etree.ElementTree as ET
from os.path import join as pjoin

import torch
from multiprocess import Pool

from .utils import _get_word_ngrams
from ..others.tokenization import BertTokenizer

nyt_remove_words = ["photo", "graph", "chart", "map", "table", "drawing"]

REMAP = {"-lrb-": "(", "-rrb-": ")", "-lcb-": "{", "-rcb-": "}",
         "-lsb-": "[", "-rsb-": "]", "``": '"', "''": '"'}


def clean(x):
    return re.sub(
        r"-lrb-|-rrb-|-lcb-|-rcb-|-lsb-|-rsb-|``|''",
        lambda m: REMAP.get(m.group()), x)


def recover_from_corenlp(s):
    s = re.sub(r' \'{\w}', r'\'\g<1>', s)
    s = re.sub(r'\'\' {\w}', r'\'\'\g<1>', s)


def load_json(p, lower):
    source = []
    tgt = []
    flag = False
    for sent in json.load(open(p))['sentences']:
        tokens = [t['word'] for t in sent['tokens']]
        if (lower):
            tokens = [t.lower() for t in tokens]
        if (tokens[0] == '@highlight'):
            flag = True
            tgt.append([])
            continue
        if (flag):
            tgt[-1].extend(tokens)
        else:
            source.append(tokens)

    source = [clean(' '.join(sent)).split() for sent in source]
    tgt = [clean(' '.join(sent)).split() for sent in tgt]
    return source, tgt


def load_xml(p):
    tree = ET.parse(p)
    root = tree.getroot()
    title, byline, abs, paras = [], [], [], []
    title_node = list(root.iter('hedline'))
    if (len(title_node) > 0):
        try:
            title = [p.text.lower().split() for p in list(title_node[0].iter('hl1'))][0]
        except:
            print(p)

    else:
        return None, None
    byline_node = list(root.iter('byline'))
    byline_node = [n for n in byline_node if n.attrib['class'] == 'normalized_byline']
    if (len(byline_node) > 0):
        byline = byline_node[0].text.lower().split()
    abs_node = list(root.iter('abstract'))
    if (len(abs_node) > 0):
        try:
            abs = [p.text.lower().split() for p in list(abs_node[0].iter('p'))][0]
        except:
            print(p)

    else:
        return None, None
    abs = ' '.join(abs).split(';')
    abs[-1] = abs[-1].replace('(m)', '')
    abs[-1] = abs[-1].replace('(s)', '')

    for ww in nyt_remove_words:
        abs[-1] = abs[-1].replace('(' + ww + ')', '')
    abs = [p.split() for p in abs]
    abs = [p for p in abs if len(p) > 2]

    for doc_node in root.iter('block'):
        att = doc_node.get('class')
        # if(att == 'abstract'):
        #     abs = [p.text for p in list(f.iter('p'))]
        if (att == 'full_text'):
            paras = [p.text.lower().split() for p in list(doc_node.iter('p'))]
            break
    if (len(paras) > 0):
        if (len(byline) > 0):
            paras = [title + ['[unused3]'] + byline + ['[unused4]']] + paras
        else:
            paras = [title + ['[unused3]']] + paras

        return paras, abs
    else:
        return None, None


def tokenize(args):
    stories_dir = os.path.abspath(args.raw_path)
    tokenized_stories_dir = os.path.abspath(args.save_path)

    print("Preparing to tokenize %s to %s..." % (stories_dir, tokenized_stories_dir))
    stories = os.listdir(stories_dir)
    # make IO list file
    print("Making list of files to tokenize...")
    with open("mapping_for_corenlp.txt", "w") as f:
        for s in stories:
            if (not s.endswith('story')):
                continue
            f.write("%s\n" % (os.path.join(stories_dir, s)))
    command = ['java', 'edu.stanford.nlp.pipeline.StanfordCoreNLP', '-annotators', 'tokenize,ssplit',
               '-ssplit.newlineIsSentenceBreak', 'always', '-filelist', 'mapping_for_corenlp.txt', '-outputFormat',
               'json', '-outputDirectory', tokenized_stories_dir]
    print("Tokenizing %i files in %s and saving in %s..." % (len(stories), stories_dir, tokenized_stories_dir))
    subprocess.call(command)
    print("Stanford CoreNLP Tokenizer has finished.")
    os.remove("mapping_for_corenlp.txt")

    # Check that the tokenized stories directory contains the same number of files as the original directory
    num_orig = len(os.listdir(stories_dir))
    num_tokenized = len(os.listdir(tokenized_stories_dir))
    if num_orig != num_tokenized:
        raise Exception(
            "The tokenized stories directory %s contains %i files, but it should contain the same number as %s (which has %i files). Was there an error during tokenization?" % (
                tokenized_stories_dir, num_tokenized, stories_dir, num_orig))
    print("Successfully finished tokenizing %s to %s.\n" % (stories_dir, tokenized_stories_dir))


def cal_rouge(evaluated_ngrams, reference_ngrams):
    reference_count = len(reference_ngrams)
    evaluated_count = len(evaluated_ngrams)

    overlapping_ngrams = evaluated_ngrams.intersection(reference_ngrams)
    overlapping_count = len(overlapping_ngrams)

    if evaluated_count == 0:
        precision = 0.0
    else:
        precision = overlapping_count / evaluated_count

    if reference_count == 0:
        recall = 0.0
    else:
        recall = overlapping_count / reference_count

    f1_score = 2.0 * ((precision * recall) / (precision + recall + 1e-8))
    return {"f": f1_score, "p": precision, "r": recall}


def greedy_selection(doc_sent_list, abstract_sent_list, summary_size):
    def _rouge_clean(s):
        return re.sub(r'[^a-zA-Z0-9 ]', '', s)

    max_rouge = 0.0
    abstract = sum(abstract_sent_list, [])
    abstract = _rouge_clean(' '.join(abstract)).split()
    sents = [_rouge_clean(' '.join(s)).split() for s in doc_sent_list]
    evaluated_1grams = [_get_word_ngrams(1, [sent]) for sent in sents]
    reference_1grams = _get_word_ngrams(1, [abstract])
    evaluated_2grams = [_get_word_ngrams(2, [sent]) for sent in sents]
    reference_2grams = _get_word_ngrams(2, [abstract])

    selected = []
    for s in range(summary_size):
        cur_max_rouge = max_rouge
        cur_id = -1
        for i in range(len(sents)):
            if (i in selected):
                continue
            c = selected + [i]
            candidates_1 = [evaluated_1grams[idx] for idx in c]
            candidates_1 = set.union(*map(set, candidates_1))
            candidates_2 = [evaluated_2grams[idx] for idx in c]
            candidates_2 = set.union(*map(set, candidates_2))
            rouge_1 = cal_rouge(candidates_1, reference_1grams)['f']
            rouge_2 = cal_rouge(candidates_2, reference_2grams)['f']
            rouge_score = rouge_1 + rouge_2
            if rouge_score > cur_max_rouge:
                cur_max_rouge = rouge_score
                cur_id = i
        if (cur_id == -1):
            return selected
        selected.append(cur_id)
        max_rouge = cur_max_rouge

    return sorted(selected)


def hashhex(s):
    """Returns a heximal formated SHA1 hash of the input string."""
    h = hashlib.sha1()
    h.update(s.encode('utf-8'))
    return h.hexdigest()


class BertData():
    def __init__(self, args):
        self.args = args
        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased', do_lower_case=True)

        self.sep_token = '[SEP]'
        self.cls_token = '[CLS]'
        self.pad_token = '[PAD]'
        self.tgt_bos = '[unused0]'
        self.tgt_eos = '[unused1]'
        self.tgt_sent_split = '[unused2]'
        self.sep_vid = self.tokenizer.vocab[self.sep_token]
        self.cls_vid = self.tokenizer.vocab[self.cls_token]
        self.pad_vid = self.tokenizer.vocab[self.pad_token]
        self.rhetorical_role_map = {'FAC': 1, 'RLC': 2, 'ARG_PETITIONER': 3, 'ARG_RESPONDENT': 4, 'ISSUE': 5,
                                    'PRE_RELIED': 6, 'PRE_NOT_RELIED': 7,
                                    'ANALYSIS': 8, 'STA': 9, 'RATIO': 10, "RPC": 11}

    def preprocess(self, src, tgt, sent_labels, sent_rhetorical_roles, use_bert_basic_tokenizer=False, is_test=False):

        if ((not is_test) and len(src) == 0):
            return None

        original_src_txt = [' '.join(s) for s in src]

        idxs = [i for i, s in enumerate(src) if (len(s) > self.args.min_src_ntokens_per_sent)]

        _sent_labels = [0] * len(src)
        for l in sent_labels:
            _sent_labels[l] = 1

        src = [src[i][:self.args.max_src_ntokens_per_sent] for i in idxs]
        sent_labels = [_sent_labels[i] for i in idxs]
        src = src[:self.args.max_src_nsents]
        sent_labels = sent_labels[:self.args.max_src_nsents]

        sent_rr = [self.rhetorical_role_map.get(i) for i in sent_rhetorical_roles]
        sent_rr = sent_rr[:self.args.max_src_nsents]

        if ((not is_test) and len(src) < self.args.min_src_nsents):
            return None

        src_txt = [' '.join(sent) for sent in src]
        text = ' {} {} '.format(self.sep_token, self.cls_token).join(src_txt)

        src_subtokens = self.tokenizer.tokenize(text)

        src_subtokens = [self.cls_token] + src_subtokens + [self.sep_token]
        src_subtoken_idxs = self.tokenizer.convert_tokens_to_ids(src_subtokens)
        _segs = [-1] + [i for i, t in enumerate(src_subtoken_idxs) if t == self.sep_vid]
        segs = [_segs[i] - _segs[i - 1] for i in range(1, len(_segs))]
        segments_ids = []
        for i, s in enumerate(segs):
            if (i % 2 == 0):
                segments_ids += s * [0]
            else:
                segments_ids += s * [1]
        cls_ids = [i for i, t in enumerate(src_subtoken_idxs) if t == self.cls_vid]
        sent_labels = sent_labels[:len(cls_ids)]
        sent_rr = sent_rr[:len(cls_ids)]

        tgt_subtokens_str = '[unused0] ' + ' [unused2] '.join(
            [' '.join(self.tokenizer.tokenize(' '.join(tt), use_bert_basic_tokenizer=use_bert_basic_tokenizer)) for tt
             in tgt]) + ' [unused1]'
        tgt_subtoken = tgt_subtokens_str.split()[:self.args.max_tgt_ntokens]
        if ((not is_test) and len(tgt_subtoken) < self.args.min_tgt_ntokens):
            return None

        tgt_subtoken_idxs = self.tokenizer.convert_tokens_to_ids(tgt_subtoken)

        tgt_txt = '<q>'.join([' '.join(tt) for tt in tgt])
        src_txt = [original_src_txt[i] for i in idxs]

        return src_subtoken_idxs, sent_labels, sent_rr, tgt_subtoken_idxs, segments_ids, cls_ids, src_txt, tgt_txt


def format_to_bert(args):
    if (args.dataset != ''):
        datasets = [args.dataset]
    else:
        datasets = ['train', 'valid', 'test']
    for corpus_type in datasets:
        a_lst = []
        for json_f in glob.glob(pjoin(args.raw_path, '*' + corpus_type + '.*.json')):
            real_name = json_f.split('/')[-1]
            a_lst.append((corpus_type, json_f, args, pjoin(args.save_path, real_name.replace('json', 'bert.pt'))))
        print(a_lst)
        pool = Pool(args.n_cpus)
        for d in pool.imap(_format_to_bert, a_lst):
            pass

        pool.close()
        pool.join()


def _format_to_bert(params):
    corpus_type, json_file, args, save_file = params
    is_test = corpus_type == 'test'
    if (os.path.exists(save_file)):
        # msg.info('Ignore %s' % save_file)
        return

    bert = BertData(args)

    # msg.info('Processing %s' % json_file)
    jobs = json.load(open(json_file))
    datasets = []
    for d in jobs:
        source, tgt = d['src'], d['tgt']
        f_name_chunk_id = '___'.join([d['source_filename'], str(d['src_chunk_id'])])
        sent_labels = greedy_selection(source[:args.max_src_nsents], tgt, len(tgt))
        sent_rhetorical_roles = d['src_rhetorical_roles']
        if (args.lower):
            source = [' '.join(s).lower().split() for s in source]
            tgt = [' '.join(s).lower().split() for s in tgt]
        b_data = bert.preprocess(source, tgt, sent_labels, sent_rhetorical_roles,
                                 use_bert_basic_tokenizer=args.use_bert_basic_tokenizer,
                                 is_test=is_test)
        # b_data = bert.preprocess(source, tgt, sent_labels, use_bert_basic_tokenizer=args.use_bert_basic_tokenizer)

        if (b_data is None):
            continue
        src_subtoken_idxs, sent_labels, sent_rr, tgt_subtoken_idxs, segments_ids, cls_ids, src_txt, tgt_txt = b_data
        b_data_dict = {"src": src_subtoken_idxs, "tgt": tgt_subtoken_idxs,
                       "src_sent_labels": sent_labels, "sentence_rhetorical_roles": sent_rr, "segs": segments_ids,
                       'clss': cls_ids,
                       'src_txt': src_txt, "tgt_txt": tgt_txt, 'unique_id': f_name_chunk_id}
        datasets.append(b_data_dict)
    # msg.info('Processed instances %d' % len(datasets))
    # msg.info('Saving to %s' % save_file)
    torch.save(datasets, save_file)
    datasets = []
    gc.collect()


def format_to_lines(args):
    corpus_mapping = {}
    for corpus_type in ['valid', 'test', 'train']:
        temp = []
        for line in open(pjoin(args.map_path, 'mapping_' + corpus_type + '.txt')):
            temp.append(hashhex(line.strip()))
        corpus_mapping[corpus_type] = {key.strip(): 1 for key in temp}
    train_files, valid_files, test_files = [], [], []
    for f in glob.glob(pjoin(args.raw_path, '*.json')):
        real_name = f.split('/')[-1].split('.')[0]
        if (real_name in corpus_mapping['valid']):
            valid_files.append(f)
        elif (real_name in corpus_mapping['test']):
            test_files.append(f)
        elif (real_name in corpus_mapping['train']):
            train_files.append(f)
        # else:
        #     train_files.append(f)

    corpora = {'train': train_files, 'valid': valid_files, 'test': test_files}
    for corpus_type in ['train', 'valid', 'test']:
        a_lst = [(f, args) for f in corpora[corpus_type]]
        pool = Pool(args.n_cpus)
        dataset = []
        p_ct = 0
        for d in pool.imap_unordered(_format_to_lines, a_lst):
            dataset.append(d)
            if (len(dataset) > args.shard_size):
                pt_file = "{:s}.{:s}.{:d}.json".format(args.save_path, corpus_type, p_ct)
                with open(pt_file, 'w') as save:
                    # save.write('\n'.join(dataset))
                    save.write(json.dumps(dataset))
                    p_ct += 1
                    dataset = []

        pool.close()
        pool.join()
        if (len(dataset) > 0):
            pt_file = "{:s}.{:s}.{:d}.json".format(args.save_path, corpus_type, p_ct)
            with open(pt_file, 'w') as save:
                # save.write('\n'.join(dataset))
                save.write(json.dumps(dataset))
                p_ct += 1
                dataset = []


def _format_to_lines(params):
    f, args = params
    print(f)
    source, tgt = load_json(f, args.lower)
    return {'src': source, 'tgt': tgt}


def format_xsum_to_lines(args):
    if (args.dataset != ''):
        datasets = [args.dataset]
    else:
        datasets = ['train', 'test', 'valid']

    corpus_mapping = json.load(open(pjoin(args.raw_path, 'XSum-TRAINING-DEV-TEST-SPLIT-90-5-5.json')))

    for corpus_type in datasets:
        mapped_fnames = corpus_mapping[corpus_type]
        root_src = pjoin(args.raw_path, 'restbody')
        root_tgt = pjoin(args.raw_path, 'firstsentence')
        # realnames = [fname.split('.')[0] for fname in os.listdir(root_src)]
        realnames = mapped_fnames

        a_lst = [(root_src, root_tgt, n) for n in realnames]
        pool = Pool(args.n_cpus)
        dataset = []
        p_ct = 0
        for d in pool.imap_unordered(_format_xsum_to_lines, a_lst):
            if (d is None):
                continue
            dataset.append(d)
            if (len(dataset) > args.shard_size):
                pt_file = "{:s}.{:s}.{:d}.json".format(args.save_path, corpus_type, p_ct)
                with open(pt_file, 'w') as save:
                    save.write(json.dumps(dataset))
                    p_ct += 1
                    dataset = []

        pool.close()
        pool.join()
        if (len(dataset) > 0):
            pt_file = "{:s}.{:s}.{:d}.json".format(args.save_path, corpus_type, p_ct)
            with open(pt_file, 'w') as save:
                save.write(json.dumps(dataset))
                p_ct += 1
                dataset = []


def _format_xsum_to_lines(params):
    src_path, root_tgt, name = params
    f_src = pjoin(src_path, name + '.restbody')
    f_tgt = pjoin(root_tgt, name + '.fs')
    if (os.path.exists(f_src) and os.path.exists(f_tgt)):
        print(name)
        source = []
        for sent in open(f_src):
            source.append(sent.split())
        tgt = []
        for sent in open(f_tgt):
            tgt.append(sent.split())
        return {'src': source, 'tgt': tgt}
    return None



import json
import os

import spacy
from tqdm import tqdm
from wasabi import msg

spacy.prefer_gpu()


def attach_short_sentence_boundries_to_next(revised_sentence_boundries, doc_txt):
    ###### this function accepts the list in the format of output of function "extract_relevant_sentences_for_rhetorical_roles" and returns the revised list with shorter sentences attached to next sentence
    min_char_cnt_per_sentence = 5

    concatenated_sentence_boundries = []
    sentences_to_attach_to_next = ()
    for sentence_boundry in revised_sentence_boundries:
        sentence_txt = doc_txt[sentence_boundry[0]: sentence_boundry[1]]
        if not sentence_txt.isspace():  ### sentences containing only spaces , newlines are discarded
            if sentences_to_attach_to_next:
                sentence_start_char = sentences_to_attach_to_next[0]
            else:
                sentence_start_char = sentence_boundry[0]
            # sentence_length_char = sentence_boundry[1] - sentence_start_char
            sentence_length_char = len(doc_txt[sentence_start_char: sentence_boundry[1]].strip())
            if sentence_length_char > min_char_cnt_per_sentence:
                concatenated_sentence_boundries.append((sentence_start_char, sentence_boundry[1]))
                sentences_to_attach_to_next = ()
            else:
                if not sentences_to_attach_to_next:
                    sentences_to_attach_to_next = sentence_boundry
    return concatenated_sentence_boundries


def split_into_sentences_tokenize_write(data, custom_processed_data_path,
                                        hsln_format_txt_dirpath='datasets/pubmed-20k', verbose=False):
    ########## This function accepts the input files in LS format, creates tokens and writes them with label as "NONE" to text file

    if not os.path.exists(hsln_format_txt_dirpath):
        os.makedirs(hsln_format_txt_dirpath)
    max_length = 10000
    output_json = []
    filename_sent_boundries = {}  ###### key is the filename and value is dict containing sentence spans {"abc.txt":{"sentence_span":[(1,10),(11,20),...]} , "pqr.txt":{...},...}
    if verbose:
        msg.info('Preprocessing rhetorical role model input!!!')
    for data_dict in tqdm(data, disable=not verbose):

        doc_id = data_dict['file_id']
        preamble_doc = data_dict['preamble_doc']
        judgment_doc = data_dict['judgement_doc']

        if filename_sent_boundries.get(doc_id) is None:  ##### Ignore if the file is already present

            nlp_doc = spacy.tokens.Doc.from_docs([preamble_doc, judgment_doc])
            doc_txt = nlp_doc.text
            sentence_boundries = [(sent.start_char, sent.end_char) for sent in nlp_doc.sents]
            revised_sentence_boundries = attach_short_sentence_boundries_to_next(sentence_boundries, doc_txt)

            adjudicated_doc = {'id': doc_id,
                               'data': {'preamble_text': preamble_doc.text,
                                        'judgement_text': judgment_doc.text,
                                        'text': doc_txt}
                               }

            adjudicated_doc['annotations'] = []
            adjudicated_doc['annotations'].append({})
            adjudicated_doc['annotations'] = []

            filename_sent_boundries[doc_id] = {"sentence_span": []}
            for sentence_boundry in revised_sentence_boundries:
                sentence_txt = doc_txt[sentence_boundry[0]:sentence_boundry[1]]

                if sentence_txt.strip() != "":
                    sentence_txt = sentence_txt.replace("\r", "")
                    sent_data = {}
                    sent_data['start'] = sentence_boundry[0]
                    sent_data['end'] = sentence_boundry[1]
                    sent_data['text'] = sentence_txt
                    sent_data['labels'] = []
                    adjudicated_doc['annotations'].append(sent_data)

        output_json.append(adjudicated_doc)
    with open(custom_processed_data_path, 'w+') as f:
        json.dump(output_json, f)


def write_in_hsln_format(input_json, hsln_format_txt_dirpath, tokenizer):
    # tokenizer = BertTokenizer.from_pretrained(BERT_VOCAB, do_lower_case=True)
    json_format = json.load(open(input_json))
    final_string = ''
    filename_sent_boundries = {}
    for file in json_format:
        file_name = file['id']
        final_string = final_string + '###' + str(file_name) + "\n"
        filename_sent_boundries[file_name] = {"sentence_span": []}
        for annotation in file['annotations']:
            filename_sent_boundries[file_name]['sentence_span'].append(
                [annotation['start'], annotation['end']])

            sentence_txt = annotation['text']
            sentence_txt = sentence_txt.replace("\r", "")
            if sentence_txt.strip() != "":
                sent_tokens = tokenizer.encode(sentence_txt, add_special_tokens=True, max_length=128)
                sent_tokens = [str(i) for i in sent_tokens]
                sent_tokens_txt = " ".join(sent_tokens)
                final_string = final_string + "NONE" + "\t" + sent_tokens_txt + "\n"
        final_string = final_string + "\n"

    with open(hsln_format_txt_dirpath + '/test_scibert.txt', "w+") as file:
        file.write(final_string)

    with open(hsln_format_txt_dirpath + '/train_scibert.txt', "w+") as file:
        file.write(final_string)
    with open(hsln_format_txt_dirpath + '/dev_scibert.txt', "w+") as file:
        file.write(final_string)
    with open(hsln_format_txt_dirpath + '/sentece_boundries.json', 'w+') as json_file:
        json.dump(filename_sent_boundries, json_file)

    return filename_sent_boundries



def reset_ids(input_json):
    """ As in result json each annotation has randon id lets set them to a common for testing purpose"""
    for i in input_json['annotations']:
        i['id'] = '1234'
        if i.get('entities') is not None:
            for j in i['entities']:
                j['id'] = '1234'
    return input_json


def reset_sent_scores(input_json):
    """ As in result json each annotation has sent_score which can change,
     so we won't test for those and reset them using this function"""
    for i in input_json['annotations']:
        i['sent_score'] = 0.0
    return input_json



import collections
import copy

import pandas as pd

from opennyai.ner.InLegalNER.InLegalNER import InLegalNER


def load(model_name: str = 'en_legal_ner_trf', use_gpu: bool = True):
    """Returns object of InLegalNER class.
     It is used for loading InLegalNER model in memory.
    Args:
        model_name (string): Accepts a model name of spacy as InLegalNER that will be used for NER inference
        available models are 'en_legal_ner_trf', 'en_legal_ner_sm'
        use_gpu (bool): Functionality to give a choice whether to use GPU for inference or not
         Setting it True doesn't ensure GPU will be utilized it need proper support libraries as mentioned in
         documentation
    """
    AVAILABLE_LEGAL_NER_MODELS = ['en_legal_ner_trf', 'en_legal_ner_sm']
    if model_name not in AVAILABLE_LEGAL_NER_MODELS:
        raise RuntimeError(f'{model_name} doesn\'t exit in list of available models {AVAILABLE_LEGAL_NER_MODELS}')
    return InLegalNER(model_name, use_gpu)


def update_json_with_provision_statute_clusters(ls_formatted_doc: dict, provision_statute_clusters: list):
    for entity, statute, normalized_provision, normalized_statute in provision_statute_clusters:
        for result in ls_formatted_doc['annotations']:
            if result['start'] == entity.start_char and result['end'] == entity.end_char:
                result['normalized_name'] = str(normalized_provision) + ' &&& ' + str(
                    normalized_statute)  ### &&& used as a separator

    return ls_formatted_doc


def update_json_with_statute_clusters(ls_formatted_doc: dict, statute_clusters: dict):
    for statute_head in statute_clusters.keys():
        for entity in statute_clusters[statute_head]:
            for result in ls_formatted_doc['annotations']:
                if result['start'] == entity.start_char and result['end'] == entity.end_char:
                    result['normalized_name'] = str(statute_head)

    return ls_formatted_doc


def update_json_with_precedent_clusters(ls_formatted_doc: dict, precedent_clusters: dict):
    for precedent_head in precedent_clusters.keys():
        for entity in precedent_clusters[precedent_head]:
            for result in ls_formatted_doc['annotations']:
                if result['start'] == entity.start_char and result['end'] == entity.end_char:
                    result['normalized_name'] = str(precedent_head)

    return ls_formatted_doc


def get_json_from_spacy_doc(doc) -> dict:
    """Returns dict of InLegalNER doc.
    Args:
        doc: InLegalNER doc
    """
    id = "LegalNER_" + doc.user_data['doc_id']
    output = {'id': id, 'annotations': [],
              'data': {'text': doc.text, 'original_text': doc.user_data['original_text'],
                       'preamble_end_char_offset': doc.user_data['preamble_end_char_offset']}}
    for ent in doc.ents:
        import uuid
        uid = uuid.uuid4()
        id = uid.hex
        output['annotations'].append(copy.deepcopy({"id": id,
                                                    "normalized_name": ent.text,
                                                    "start": ent.start_char,
                                                    "end": ent.end_char,
                                                    "text": ent.text,
                                                    "labels": [ent.label_]}))

    if doc.user_data.get('provision_statute_pairs') is not None:
        output = update_json_with_provision_statute_clusters(copy.deepcopy(output),
                                                             doc.user_data['provision_statute_pairs'])
    if doc.user_data.get('statute_clusters') is not None:
        output = update_json_with_statute_clusters(copy.deepcopy(output), doc.user_data['statute_clusters'])
    if doc.user_data.get('precedent_clusters') is not None:
        output = update_json_with_precedent_clusters(copy.deepcopy(output),
                                                     doc.user_data['precedent_clusters'])

        return output
    else:
        return output


def get_csv(doc, f_name, save_path):
    df = pd.DataFrame(columns=['file_name', 'entity', 'label', 'normalised_entities'])
    file_name = []
    entity = []
    label = []
    normalised_entities = []
    if doc.user_data.get('provision_statute_pairs') is None:
        doc.user_data['provision_statute_pairs'] = []
    if doc.user_data.get('precedent_clusters') is None:
        doc.user_data['precedent_clusters'] = {}
    if doc.user_data.get('statute_clusters') is None:
        doc.user_data['statute_clusters'] = {}
    for pro_ent in doc.user_data['provision_statute_pairs']:
        file_name.append(f_name)
        entity.append(pro_ent[0])
        label.append('PROVISION')
        normalised_entities.append(pro_ent[2] + ' of ' + pro_ent[3])
    for pre_head in doc.user_data['precedent_clusters'].keys():

        for ent in doc.user_data['precedent_clusters'][pre_head]:
            file_name.append(f_name)
            entity.append(ent)
            label.append('PRECEDENT')
            normalised_entities.append(pre_head)
    for pre_head in doc.user_data['statute_clusters'].keys():

        for ent in doc.user_data['statute_clusters'][pre_head]:
            file_name.append(f_name)
            entity.append(ent)
            label.append('STATUTE')
            normalised_entities.append(pre_head)

    for ent in doc.ents:
        if ent not in entity:
            file_name.append(f_name)
            entity.append(ent)
            label.append(ent.label_)
            normalised_entities.append('')
    entity_text = [ent.text for ent in entity]
    df['file_name'] = file_name
    df['entity'] = entity_text
    df['label'] = label
    df['normalised_entities'] = normalised_entities

    df.to_csv(save_path)


def get_unique_precedent_count(doc):
    new_clusters = {}
    clusters = doc.user_data['precedent_clusters']
    for c in clusters.keys():
        new_clusters[c] = len(clusters[c])

    return new_clusters


def get_unique_provision_count(doc):
    clusters = doc.user_data['provision_statute_pairs']
    provisions = [cluster[2] + ' of ' + cluster[3] for cluster in clusters]
    frequency = dict(collections.Counter(provisions))

    return frequency


def get_unique_statute_count(doc):
    clusters = doc.user_data['provision_statute_pairs']
    statutes = [cluster[3] for cluster in clusters]
    frequency = dict(collections.Counter(statutes))

    return frequency


ner_displacy_option = {
    'colors': {'COURT': "#bbabf2", 'PETITIONER': "#f570ea", "RESPONDENT": "#cdee81", 'JUDGE': "#fdd8a5",
               "LAWYER": "#f9d380", 'WITNESS': "violet", "STATUTE": "#faea99", "PROVISION": "yellow",
               'CASE_NUMBER': "#fbb1cf", "PRECEDENT": "#fad6d6", 'DATE': "#b1ecf7", 'OTHER_PERSON': "#b0f6a2",
               'ORG': '#a57db5', 'GPE': '#7fdbd4'}}



import torch
from tqdm import tqdm
from wasabi import msg

from .utils import tensor_dict_to_gpu, tensor_dict_to_cpu


def infer_model(model, eval_batches, device, task, verbose=False):
    model.eval()
    labels_dict = {}
    predicted_labels = []
    doc_name_list = []
    docwise_predicted_labels = []
    if verbose:
        msg.info('Processing documents with rhetorical role model!!!')
    with torch.no_grad():
        for batch in tqdm(eval_batches, disable=not verbose):
            # move tensor to gpu
            tensor_dict_to_gpu(batch, device)

            if batch["task"] != task.task_name:
                continue

            output = model(batch=batch)

            true_labels_batch, predicted_labels_batch = \
                clear_and_map_padded_values(batch["label_ids"].view(-1), output["predicted_label"].view(-1),
                                            task.labels)
            doc_name_list.append(batch['doc_name'][0])
            docwise_predicted_labels.append(predicted_labels_batch)

            predicted_labels.extend(predicted_labels_batch)
            tensor_dict_to_cpu(batch)

    labels_dict['y_predicted'] = predicted_labels
    labels_dict['labels'] = task.labels
    labels_dict['doc_names'] = doc_name_list
    labels_dict['docwise_y_predicted'] = docwise_predicted_labels

    return labels_dict


def clear_and_map_padded_values(true_labels, predicted_labels, labels):
    assert len(true_labels) == len(predicted_labels)
    cleared_predicted = []
    cleared_true = []
    for true_label, predicted_label in zip(true_labels, predicted_labels):
        # filter masked labels (0)
        if true_label > 0:
            cleared_true.append(labels[true_label])
            cleared_predicted.append(labels[predicted_label])
    return cleared_true, cleared_predicted



import bisect
import gc
import glob
import random

import torch


class Batch(object):
    def _pad(self, data, pad_id, width=-1):
        if (width == -1):
            width = max(len(d) for d in data)
        rtn_data = [d + [pad_id] * (width - len(d)) for d in data]
        return rtn_data

    def __init__(self, data=None, device=None, is_test=False):
        """Create a Batch from a list of examples."""
        if data is not None:
            self.batch_size = len(data)
            pre_src = [x[0] for x in data]
            pre_tgt = [x[1] for x in data]
            pre_segs = [x[2] for x in data]
            pre_clss = [x[3] for x in data]
            pre_src_sent_labels = [x[4] for x in data]
            pre_sent_rhetorical_roles = [x[5] for x in data]
            #             self.unique_ids = [x[7] for x in data]

            src = torch.tensor(self._pad(pre_src, 0))
            tgt = torch.tensor(self._pad(pre_tgt, 0))

            segs = torch.tensor(self._pad(pre_segs, 0))
            mask_src = ~ (src == 0)
            mask_tgt = ~ (tgt == 0)

            clss = torch.tensor(self._pad(pre_clss, -1))
            src_sent_labels = torch.tensor(self._pad(pre_src_sent_labels, 0))
            sent_rhetorical_roles = torch.tensor(self._pad(pre_sent_rhetorical_roles, 0))
            mask_cls = ~ (clss == -1)
            clss[clss == -1] = 0
            setattr(self, 'clss', clss.to(device))
            setattr(self, 'mask_cls', mask_cls.to(device))
            setattr(self, 'src_sent_labels', src_sent_labels.to(device))
            setattr(self, 'sentence_rhetorical_roles', sent_rhetorical_roles.to(device))

            setattr(self, 'src', src.to(device))
            setattr(self, 'tgt', tgt.to(device))
            setattr(self, 'segs', segs.to(device))
            setattr(self, 'mask_src', mask_src.to(device))
            setattr(self, 'mask_tgt', mask_tgt.to(device))

            if (is_test):
                src_str = [x[-3] for x in data]
                setattr(self, 'src_str', src_str)
                tgt_str = [x[-2] for x in data]
                setattr(self, 'tgt_str', tgt_str)
                unique_ids = [x[-1] for x in data]
                setattr(self, 'unique_ids', unique_ids)

    def __len__(self):
        return self.batch_size


def load_dataset(args, corpus_type, shuffle):
    """
    Dataset generator. Don't do extra stuff here, like printing,
    because they will be postponed to the first loading time.

    Args:
        corpus_type: 'train' or 'valid'
    Returns:
        A list of dataset, the dataset(s) are lazily loaded.
    """
    assert corpus_type in ["train", "valid", "test"]

    def _lazy_dataset_loader(pt_file, corpus_type):
        dataset = torch.load(pt_file)
        return dataset

    # Sort the glob output by file name (by increasing indexes).
    pts = sorted(glob.glob(args.bert_data_path + '.' + corpus_type + '.[0-9]*.pt'))
    if pts:
        if (shuffle):
            random.shuffle(pts)

        for pt in pts:
            yield _lazy_dataset_loader(pt, corpus_type)
    else:
        # Only one inputters.*Dataset, simple!
        pt = args.bert_data_path + '.' + corpus_type + '.pt'
        yield _lazy_dataset_loader(pt, corpus_type)


def abs_batch_size_fn(new, count):
    src, tgt = new[0], new[1]
    global max_n_sents, max_n_tokens, max_size
    if count == 1:
        max_size = 0
        max_n_sents = 0
        max_n_tokens = 0
    max_n_sents = max(max_n_sents, len(tgt))
    max_size = max(max_size, max_n_sents)
    src_elements = count * max_size
    if (count > 6):
        return src_elements + 1e3
    return src_elements


def ext_batch_size_fn(new, count):
    if (len(new) == 4):
        pass
    src, labels = new[0], new[4]
    global max_n_sents, max_n_tokens, max_size
    if count == 1:
        max_size = 0
        max_n_sents = 0
        max_n_tokens = 0
    max_n_sents = max(max_n_sents, len(src))
    max_size = max(max_size, max_n_sents)
    src_elements = count * max_size
    return src_elements


class Dataloader(object):
    def __init__(self, args, datasets, batch_size,
                 device, shuffle, is_test):
        self.args = args
        self.datasets = datasets
        self.batch_size = batch_size
        self.device = device
        self.shuffle = shuffle
        self.is_test = is_test
        self.cur_iter = self._next_dataset_iterator(datasets)
        assert self.cur_iter is not None

    def __iter__(self):
        dataset_iter = (d for d in self.datasets)
        while self.cur_iter is not None:
            for batch in self.cur_iter:
                yield batch
            self.cur_iter = self._next_dataset_iterator(dataset_iter)

    def _next_dataset_iterator(self, dataset_iter):
        try:
            # Drop the current dataset for decreasing memory
            if hasattr(self, "cur_dataset"):
                self.cur_dataset = None
                gc.collect()
                del self.cur_dataset
                gc.collect()

            self.cur_dataset = next(dataset_iter)
        except StopIteration:
            return None

        return DataIterator(args=self.args,
                            dataset=self.cur_dataset, batch_size=self.batch_size,
                            device=self.device, shuffle=self.shuffle, is_test=self.is_test)


class DataIterator(object):
    def __init__(self, args, dataset, batch_size, device=None, is_test=False,
                 shuffle=True):
        self.args = args
        self.batch_size, self.is_test, self.dataset = batch_size, is_test, dataset
        self.iterations = 0
        self.device = device
        self.shuffle = shuffle

        self.sort_key = lambda x: len(x[1])

        self._iterations_this_epoch = 0
        if (self.args.task == 'abs'):
            self.batch_size_fn = abs_batch_size_fn
        else:
            self.batch_size_fn = ext_batch_size_fn

    def data(self):
        if self.shuffle:
            random.shuffle(self.dataset)
        xs = self.dataset
        return xs

    def preprocess(self, ex, is_test):
        src = ex['src']
        tgt = ex['tgt'][:self.args.max_tgt_len][:-1] + [2]
        src_sent_labels = ex['src_sent_labels']
        segs = ex['segs']
        if (not self.args.use_interval):
            segs = [0] * len(segs)
        clss = ex['clss']
        src_txt = ex['src_txt']
        tgt_txt = ex['tgt_txt']
        unique_id = ex['unique_id']
        sentence_rhetorical_roles = ex['sentence_rhetorical_roles']

        end_id = [src[-1]]
        src = src[:-1][:self.args.max_pos - 1] + end_id
        segs = segs[:self.args.max_pos]
        max_sent_id = bisect.bisect_left(clss, self.args.max_pos)
        src_sent_labels = src_sent_labels[:max_sent_id]
        clss = clss[:max_sent_id]
        sentence_rhetorical_roles = sentence_rhetorical_roles[:max_sent_id]
        # src_txt = src_txt[:max_sent_id]

        if (is_test):
            return src, tgt, segs, clss, src_sent_labels, sentence_rhetorical_roles, src_txt, tgt_txt, unique_id
        else:
            return src, tgt, segs, clss, src_sent_labels, sentence_rhetorical_roles, unique_id

    def batch_buffer(self, data, batch_size):
        minibatch, size_so_far = [], 0
        for ex in data:
            if (len(ex['src']) == 0):
                continue
            ex = self.preprocess(ex, self.is_test)
            if (ex is None):
                continue
            minibatch.append(ex)
            size_so_far = self.batch_size_fn(ex, len(minibatch))
            if size_so_far == batch_size:
                yield minibatch
                minibatch, size_so_far = [], 0
            elif size_so_far > batch_size:
                yield minibatch[:-1]
                minibatch, size_so_far = minibatch[-1:], self.batch_size_fn(ex, 1)
        if minibatch:
            yield minibatch

    def batch(self, data, batch_size):
        """Yield elements from data in chunks of batch_size."""
        minibatch, size_so_far = [], 0
        for ex in data:
            minibatch.append(ex)
            size_so_far = self.batch_size_fn(ex, len(minibatch))
            if size_so_far == batch_size:
                yield minibatch
                minibatch, size_so_far = [], 0
            elif size_so_far > batch_size:
                yield minibatch[:-1]
                minibatch, size_so_far = minibatch[-1:], self.batch_size_fn(ex, 1)
        if minibatch:
            yield minibatch

    def create_batches(self):
        """ Create batches """
        data = self.data()
        for buffer in self.batch_buffer(data, self.batch_size * 300):

            if (self.args.task == 'abs'):
                p_batch = sorted(buffer, key=lambda x: len(x[2]))
                p_batch = sorted(p_batch, key=lambda x: len(x[1]))
            else:
                p_batch = sorted(buffer, key=lambda x: len(x[2]))

            p_batch = self.batch(p_batch, self.batch_size)

            p_batch = list(p_batch)
            if (self.shuffle):
                random.shuffle(p_batch)
            for b in p_batch:
                if (len(b) == 0):
                    continue
                yield b

    def __iter__(self):
        while True:
            self.batches = self.create_batches()
            for idx, minibatch in enumerate(self.batches):
                # fast-forward if loaded from state
                if self._iterations_this_epoch > idx:
                    continue
                self.iterations += 1
                self._iterations_this_epoch += 1
                batch = Batch(minibatch, self.device, self.is_test)

                yield batch
            return

# class TextDataloader(object):
#     def __init__(self, args, datasets, batch_size,
#                  device, shuffle, is_test):
#         self.args = args
#         self.batch_size = batch_size
#         self.device = device
#
#     def data(self):
#         if self.shuffle:
#             random.shuffle(self.dataset)
#         xs = self.dataset
#         return xs
#
#     def preprocess(self, ex, is_test):
#         src = ex['src']
#         tgt = ex['tgt'][:self.args.max_tgt_len][:-1] + [2]
#         src_sent_labels = ex['src_sent_labels']
#         segs = ex['segs']
#         if (not self.args.use_interval):
#             segs = [0] * len(segs)
#         clss = ex['clss']
#         src_txt = ex['src_txt']
#         tgt_txt = ex['tgt_txt']
#
#         end_id = [src[-1]]
#         src = src[:-1][:self.args.max_pos - 1] + end_id
#         segs = segs[:self.args.max_pos]
#         max_sent_id = bisect.bisect_left(clss, self.args.max_pos)
#         src_sent_labels = src_sent_labels[:max_sent_id]
#         clss = clss[:max_sent_id]
#         # src_txt = src_txt[:max_sent_id]
#
#         if (is_test):
#             return src, tgt, segs, clss, src_sent_labels, src_txt, tgt_txt
#         else:
#             return src, tgt, segs, clss, src_sent_labels
#
#     def batch_buffer(self, data, batch_size):
#         minibatch, size_so_far = [], 0
#         for ex in data:
#             if (len(ex['src']) == 0):
#                 continue
#             ex = self.preprocess(ex, self.is_test)
#             if (ex is None):
#                 continue
#             minibatch.append(ex)
#             size_so_far = simple_batch_size_fn(ex, len(minibatch))
#             if size_so_far == batch_size:
#                 yield minibatch
#                 minibatch, size_so_far = [], 0
#             elif size_so_far > batch_size:
#                 yield minibatch[:-1]
#                 minibatch, size_so_far = minibatch[-1:], simple_batch_size_fn(ex, 1)
#         if minibatch:
#             yield minibatch
#
#     def create_batches(self):
#         """ Create batches """
#         data = self.data()
#         for buffer in self.batch_buffer(data, self.batch_size * 300):
#             if (self.args.task == 'abs'):
#                 p_batch = sorted(buffer, key=lambda x: len(x[2]))
#                 p_batch = sorted(p_batch, key=lambda x: len(x[1]))
#             else:
#                 p_batch = sorted(buffer, key=lambda x: len(x[2]))
#                 p_batch = batch(p_batch, self.batch_size)
#
#             p_batch = batch(p_batch, self.batch_size)
#
#             p_batch = list(p_batch)
#             if (self.shuffle):
#                 random.shuffle(p_batch)
#             for b in p_batch:
#                 if (len(b) == 0):
#                     continue
#                 yield b
#
#     def __iter__(self):
#         while True:
#             self.batches = self.create_batches()
#             for idx, minibatch in enumerate(self.batches):
#                 # fast-forward if loaded from state
#                 if self._iterations_this_epoch > idx:
#                     continue
#                 self.iterations += 1
#                 self._iterations_this_epoch += 1
#                 batch = Batch(minibatch, self.device, self.is_test)
#
#                 yield batch
#             return



"""
Conditional random field with emission- and transition-based weighting
"""
from typing import List, Tuple

import torch

from opennyai.rhetorical_roles.allennlp_helper.common.checks import ConfigurationError
from opennyai.rhetorical_roles.allennlp_helper.modules.conditional_random_field.conditional_random_field import (
    ConditionalRandomField,
)


class ConditionalRandomFieldWeightTrans(ConditionalRandomField):
    """
    This module uses the "forward-backward" algorithm to compute
    the log-likelihood of its inputs assuming a conditional random field model.

    See, e.g. http://www.cs.columbia.edu/~mcollins/fb.pdf

    This is a weighted version of `ConditionalRandomField` which accepts a `label_weights`
    parameter to be used in the loss function in order to give different weights for each
    token depending on its label. The method implemented here is based on the simple idea
    of weighting emission and transition scores using the weight given for the
    corresponding tag.

    There are two other sample weighting methods implemented. You can find more details
    about them in: https://eraldoluis.github.io/2022/05/10/weighted-crf.html

    # Parameters

    num_tags : `int`, required
        The number of tags.
    label_weights : `List[float]`, required
        A list of weights to be used in the loss function in order to
        give different weights for each token depending on its label.
        `len(label_weights)` must be equal to `num_tags`. This is useful to
        deal with highly unbalanced datasets. The method implemented here is
        based on the simple idea of weighting emission and transition scores
        using the weight given for the corresponding tag.
    constraints : `List[Tuple[int, int]]`, optional (default = `None`)
        An optional list of allowed transitions (from_tag_id, to_tag_id).
        These are applied to `viterbi_tags()` but do not affect `forward()`.
        These should be derived from `allowed_transitions` so that the
        start and end transitions are handled correctly for your tag type.
    include_start_end_transitions : `bool`, optional (default = `True`)
        Whether to include the start and end transition parameters.
    """

    def __init__(
            self,
            num_tags: int,
            label_weights: List[float],
            constraints: List[Tuple[int, int]] = None,
            include_start_end_transitions: bool = True,
    ) -> None:
        super().__init__(num_tags, constraints, include_start_end_transitions)

        if label_weights is None:
            raise ConfigurationError("label_weights must be given")

        self.register_buffer("label_weights", torch.Tensor(label_weights))

    def forward(
            self, inputs: torch.Tensor, tags: torch.Tensor, mask: torch.BoolTensor = None
    ) -> torch.Tensor:
        """Computes the log likelihood for the given batch of input sequences $(x,y)$

        Args:
            inputs (torch.Tensor): (batch_size, sequence_length, num_tags) tensor of logits for the inputs $x$
            tags (torch.Tensor): (batch_size, sequence_length) tensor of tags $y$
            mask (torch.BoolTensor, optional): (batch_size, sequence_length) tensor of masking flags.
                Defaults to None.

        Returns:
            torch.Tensor: (batch_size,) log likelihoods $log P(y|x)$ for each input
        """
        if mask is None:
            mask = torch.ones(*tags.size(), dtype=torch.bool, device=inputs.device)
        else:
            # The code below fails in weird ways if this isn't a bool tensor, so we make sure.
            mask = mask.to(torch.bool)

        label_weights = self.label_weights

        # weight transition score using current_tag, i.e., t(i,j) will be t(i,j)*w(i),
        # where t(i,j) is the score of transitioning from i to j and w(i) is the weight
        # for tag i.
        transitions = self.transitions * label_weights.view(-1, 1)

        # scale the logits for all examples and all time steps
        inputs = inputs * label_weights.view(1, 1, -1)

        log_denominator = self._input_likelihood(inputs, transitions, mask)
        log_numerator = self._joint_likelihood(inputs, transitions, tags, mask)

        return torch.sum(log_numerator - log_denominator)



import math

import torch
import torch.nn as nn
import torch.nn.functional as F


def aeq(*args):
    """
    Assert all arguments have the same value
    """
    arguments = (arg for arg in args)
    first = next(arguments)
    assert all(arg == first for arg in arguments), \
        "Not all arguments have the same value: " + str(args)


def sequence_mask(lengths, max_len=None):
    """
    Creates a boolean mask from sequence lengths.
    """
    batch_size = lengths.numel()
    max_len = max_len or lengths.max()
    return (torch.arange(0, max_len)
            .type_as(lengths)
            .repeat(batch_size, 1)
            .lt(lengths.unsqueeze(1)))


def gelu(x):
    return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))


""" Global attention modules (Luong / Bahdanau) """


class GlobalAttention(nn.Module):
    """
    Global attention takes a matrix and a query vector. It
    then computes a parameterized convex combination of the matrix
    based on the input query.

    Constructs a unit mapping a query `q` of size `dim`
    and a source matrix `H` of size `n x dim`, to an output
    of size `dim`.


    .. mermaid::

       graph BT
          A[Query]
          subgraph RNN
            C[H 1]
            D[H 2]
            E[H N]
          end
          F[Attn]
          G[Output]
          A --> F
          C --> F
          D --> F
          E --> F
          C -.-> G
          D -.-> G
          E -.-> G
          F --> G

    All models compute the output as
    :math:`c = sum_{j=1}^{SeqLength} a_j H_j` where
    :math:`a_j` is the softmax of a score function.
    Then then apply a projection layer to [q, c].

    However they
    differ on how they compute the attention score.

    * Luong Attention (dot, general):
       * dot: :math:`score(H_j,q) = H_j^T q`
       * general: :math:`score(H_j, q) = H_j^T W_a q`


    * Bahdanau Attention (mlp):
       * :math:`score(H_j, q) = v_a^T tanh(W_a q + U_a h_j)`


    Args:
       dim (int): dimensionality of query and key
       coverage (bool): use coverage term
       attn_type (str): type of attention to use, options [dot,general,mlp]

    """

    def __init__(self, dim, attn_type="dot"):
        super(GlobalAttention, self).__init__()

        self.dim = dim
        assert attn_type in ["dot", "general", "mlp"], (
            "Please select a valid attention type.")
        self.attn_type = attn_type

        if self.attn_type == "general":
            self.linear_in = nn.Linear(dim, dim, bias=False)
        elif self.attn_type == "mlp":
            self.linear_context = nn.Linear(dim, dim, bias=False)
            self.linear_query = nn.Linear(dim, dim, bias=True)
            self.v = nn.Linear(dim, 1, bias=False)
        # mlp wants it with bias
        out_bias = self.attn_type == "mlp"
        self.linear_out = nn.Linear(dim * 2, dim, bias=out_bias)

    def score(self, h_t, h_s):
        """
        Args:
          h_t (`FloatTensor`): sequence of queries `[batch x tgt_len x dim]`
          h_s (`FloatTensor`): sequence of sources `[batch x src_len x dim]`

        Returns:
          :obj:`FloatTensor`:
           raw attention scores (unnormalized) for each src index
          `[batch x tgt_len x src_len]`

        """

        # Check input sizes
        src_batch, src_len, src_dim = h_s.size()
        tgt_batch, tgt_len, tgt_dim = h_t.size()

        if self.attn_type in ["general", "dot"]:
            if self.attn_type == "general":
                h_t_ = h_t.view(tgt_batch * tgt_len, tgt_dim)
                h_t_ = self.linear_in(h_t_)
                h_t = h_t_.view(tgt_batch, tgt_len, tgt_dim)
            h_s_ = h_s.transpose(1, 2)
            # (batch, t_len, d) x (batch, d, s_len) --> (batch, t_len, s_len)
            return torch.bmm(h_t, h_s_)
        else:
            dim = self.dim
            wq = self.linear_query(h_t.view(-1, dim))
            wq = wq.view(tgt_batch, tgt_len, 1, dim)
            wq = wq.expand(tgt_batch, tgt_len, src_len, dim)

            uh = self.linear_context(h_s.contiguous().view(-1, dim))
            uh = uh.view(src_batch, 1, src_len, dim)
            uh = uh.expand(src_batch, tgt_len, src_len, dim)

            # (batch, t_len, s_len, d)
            wquh = torch.tanh(wq + uh)

            return self.v(wquh.view(-1, dim)).view(tgt_batch, tgt_len, src_len)

    def forward(self, source, memory_bank, memory_lengths=None, memory_masks=None):
        """

        Args:
          source (`FloatTensor`): query vectors `[batch x tgt_len x dim]`
          memory_bank (`FloatTensor`): source vectors `[batch x src_len x dim]`
          memory_lengths (`LongTensor`): the source context lengths `[batch]`
          coverage (`FloatTensor`): None (not supported yet)

        Returns:
          (`FloatTensor`, `FloatTensor`):

          * Computed vector `[tgt_len x batch x dim]`
          * Attention distribtutions for each query
             `[tgt_len x batch x src_len]`
        """

        # one step input
        if source.dim() == 2:
            one_step = True
            source = source.unsqueeze(1)
        else:
            one_step = False

        batch, source_l, dim = memory_bank.size()
        batch_, target_l, dim_ = source.size()

        # compute attention scores, as in Luong et al.
        align = self.score(source, memory_bank)

        if memory_masks is not None:
            memory_masks = memory_masks.transpose(0, 1)
            memory_masks = memory_masks.transpose(1, 2)
            align.masked_fill_(1 - memory_masks.byte(), -float('inf'))

        if memory_lengths is not None:
            mask = sequence_mask(memory_lengths, max_len=align.size(-1))
            mask = mask.unsqueeze(1)  # Make it broadcastable.
            align.masked_fill_(1 - mask, -float('inf'))

        align_vectors = F.softmax(align.view(batch * target_l, source_l), -1)
        align_vectors = align_vectors.view(batch, target_l, source_l)

        c = torch.bmm(align_vectors, memory_bank)

        # concatenate
        concat_c = torch.cat([c, source], 2).view(batch * target_l, dim * 2)
        attn_h = self.linear_out(concat_c).view(batch, target_l, dim)
        if self.attn_type in ["general", "dot"]:
            attn_h = torch.tanh(attn_h)

        if one_step:
            attn_h = attn_h.squeeze(1)
            align_vectors = align_vectors.squeeze(1)


        else:
            attn_h = attn_h.transpose(0, 1).contiguous()
            align_vectors = align_vectors.transpose(0, 1).contiguous()

        return attn_h, align_vectors


class PositionwiseFeedForward(nn.Module):
    """ A two-layer Feed-Forward-Network with residual layer norm.

    Args:
        d_model (int): the size of input for the first-layer of the FFN.
        d_ff (int): the hidden layer size of the second-layer
            of the FNN.
        dropout (float): dropout probability in :math:`[0, 1)`.
    """

    def __init__(self, d_model, d_ff, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)
        self.actv = gelu
        self.dropout_1 = nn.Dropout(dropout)
        self.dropout_2 = nn.Dropout(dropout)

    def forward(self, x):
        inter = self.dropout_1(self.actv(self.w_1(self.layer_norm(x))))
        output = self.dropout_2(self.w_2(inter))
        return output + x


class MultiHeadedAttention(nn.Module):
    """
    Multi-Head Attention module from
    "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`.

    Similar to standard `dot` attention but uses
    multiple attention distributions simulataneously
    to select relevant items.

    .. mermaid::

       graph BT
          A[key]
          B[value]
          C[query]
          O[output]
          subgraph Attn
            D[Attn 1]
            E[Attn 2]
            F[Attn N]
          end
          A --> D
          C --> D
          A --> E
          C --> E
          A --> F
          C --> F
          D --> O
          E --> O
          F --> O
          B --> O

    Also includes several additional tricks.

    Args:
       head_count (int): number of parallel heads
       model_dim (int): the dimension of keys/values/queries,
           must be divisible by head_count
       dropout (float): dropout parameter
    """

    def __init__(self, head_count, model_dim, dropout=0.1, use_final_linear=True):
        assert model_dim % head_count == 0
        self.dim_per_head = model_dim // head_count
        self.model_dim = model_dim

        super(MultiHeadedAttention, self).__init__()
        self.head_count = head_count

        self.linear_keys = nn.Linear(model_dim,
                                     head_count * self.dim_per_head)
        self.linear_values = nn.Linear(model_dim,
                                       head_count * self.dim_per_head)
        self.linear_query = nn.Linear(model_dim,
                                      head_count * self.dim_per_head)
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.use_final_linear = use_final_linear
        if (self.use_final_linear):
            self.final_linear = nn.Linear(model_dim, model_dim)

    def forward(self, key, value, query, mask=None,
                layer_cache=None, type=None, predefined_graph_1=None):
        """
        Compute the context vector and the attention vectors.

        Args:
           key (`FloatTensor`): set of `key_len`
                key vectors `[batch, key_len, dim]`
           value (`FloatTensor`): set of `key_len`
                value vectors `[batch, key_len, dim]`
           query (`FloatTensor`): set of `query_len`
                 query vectors  `[batch, query_len, dim]`
           mask: binary mask indicating which keys have
                 non-zero attention `[batch, query_len, key_len]`
        Returns:
           (`FloatTensor`, `FloatTensor`) :

           * output context vectors `[batch, query_len, dim]`
           * one of the attention vectors `[batch, query_len, key_len]`
        """

        # CHECKS
        # batch, k_len, d = key.size()
        # batch_, k_len_, d_ = value.size()
        # aeq(batch, batch_)
        # aeq(k_len, k_len_)
        # aeq(d, d_)
        # batch_, q_len, d_ = query.size()
        # aeq(batch, batch_)
        # aeq(d, d_)
        # aeq(self.model_dim % 8, 0)
        # if mask is not None:
        #    batch_, q_len_, k_len_ = mask.size()
        #    aeq(batch_, batch)
        #    aeq(k_len_, k_len)
        #    aeq(q_len_ == q_len)
        # END CHECKS

        batch_size = key.size(0)
        dim_per_head = self.dim_per_head
        head_count = self.head_count
        key_len = key.size(1)
        query_len = query.size(1)

        def shape(x):
            """  projection """
            return x.view(batch_size, -1, head_count, dim_per_head) \
                .transpose(1, 2)

        def unshape(x):
            """  compute context """
            return x.transpose(1, 2).contiguous() \
                .view(batch_size, -1, head_count * dim_per_head)

        # 1) Project key, value, and query.
        if layer_cache is not None:
            if type == "self":
                query, key, value = self.linear_query(query), \
                                    self.linear_keys(query), \
                                    self.linear_values(query)

                key = shape(key)
                value = shape(value)

                if layer_cache is not None:
                    device = key.device
                    if layer_cache["self_keys"] is not None:
                        key = torch.cat(
                            (layer_cache["self_keys"].to(device), key),
                            dim=2)
                    if layer_cache["self_values"] is not None:
                        value = torch.cat(
                            (layer_cache["self_values"].to(device), value),
                            dim=2)
                    layer_cache["self_keys"] = key
                    layer_cache["self_values"] = value
            elif type == "context":
                query = self.linear_query(query)
                if layer_cache is not None:
                    if layer_cache["memory_keys"] is None:
                        key, value = self.linear_keys(key), \
                                     self.linear_values(value)
                        key = shape(key)
                        value = shape(value)
                    else:
                        key, value = layer_cache["memory_keys"], \
                                     layer_cache["memory_values"]
                    layer_cache["memory_keys"] = key
                    layer_cache["memory_values"] = value
                else:
                    key, value = self.linear_keys(key), \
                                 self.linear_values(value)
                    key = shape(key)
                    value = shape(value)
        else:
            key = self.linear_keys(key)
            value = self.linear_values(value)
            query = self.linear_query(query)
            key = shape(key)
            value = shape(value)

        query = shape(query)

        key_len = key.size(2)
        query_len = query.size(2)

        # 2) Calculate and scale scores.
        query = query / math.sqrt(dim_per_head)
        scores = torch.matmul(query, key.transpose(2, 3))

        if mask is not None:
            mask = mask.unsqueeze(1).expand_as(scores)
            scores = scores.masked_fill(mask, -1e18)

        # 3) Apply attention dropout and compute context vectors.

        attn = self.softmax(scores)

        if (not predefined_graph_1 is None):
            attn_masked = attn[:, -1] * predefined_graph_1
            attn_masked = attn_masked / (torch.sum(attn_masked, 2).unsqueeze(2) + 1e-9)

            attn = torch.cat([attn[:, :-1], attn_masked.unsqueeze(1)], 1)

        drop_attn = self.dropout(attn)
        if (self.use_final_linear):
            context = unshape(torch.matmul(drop_attn, value))
            output = self.final_linear(context)
            return output
        else:
            context = torch.matmul(drop_attn, value)
            return context

        # CHECK
        # batch_, q_len_, d_ = output.size()
        # aeq(q_len, q_len_)
        # aeq(batch, batch_)
        # aeq(d, d_)

        # Return one attn


class DecoderState(object):
    """Interface for grouping together the current state of a recurrent
    decoder. In the simplest case just represents the hidden state of
    the model.  But can also be used for implementing various forms of
    input_feeding and non-recurrent models.

    Modules need to implement this to utilize beam search decoding.
    """

    def detach(self):
        """ Need to document this """
        self.hidden = tuple([_.detach() for _ in self.hidden])
        self.input_feed = self.input_feed.detach()

    def beam_update(self, idx, positions, beam_size):
        """ Need to document this """
        for e in self._all:
            sizes = e.size()
            br = sizes[1]
            if len(sizes) == 3:
                sent_states = e.view(sizes[0], beam_size, br // beam_size,
                                     sizes[2])[:, :, idx]
            else:
                sent_states = e.view(sizes[0], beam_size,
                                     br // beam_size,
                                     sizes[2],
                                     sizes[3])[:, :, idx]

            sent_states.data.copy_(
                sent_states.data.index_select(1, positions))

    def map_batch_fn(self, fn):
        raise NotImplementedError()



# stopwords = pkgutil.get_data(__package__, 'smart_common_words.txt')
# stopwords = stopwords.decode('ascii').split('\n')
# stopwords = {key.strip(): 1 for key in stopwords}


def _get_ngrams(n, text):
    """Calcualtes n-grams.

    Args:
      n: which n-grams to calculate
      text: An array of tokens

    Returns:
      A set of n-grams
    """
    ngram_set = set()
    text_length = len(text)
    max_index_ngram_start = text_length - n
    for i in range(max_index_ngram_start + 1):
        ngram_set.add(tuple(text[i:i + n]))
    return ngram_set


def _get_word_ngrams(n, sentences):
    """Calculates word n-grams for multiple sentences.
    """
    assert len(sentences) > 0
    assert n > 0

    # words = _split_into_words(sentences)

    words = sum(sentences, [])
    # words = [w for w in words if w not in stopwords]
    return _get_ngrams(n, words)



import ast
import copy

import numpy as np
import torch
from spacy.lang.en import English
from tqdm import tqdm
from transformers import BertTokenizer
from wasabi import msg

from opennyai.utils.download import load_model_from_cache
from .models import data_loader
from .models.model_builder import ExtSummarizer
from .others.args import __setargs__
from .others.postprocessing_utils import _postprocess
from .others.utils import preprocess_for_summarization, format_to_bert


class ExtractiveSummarizer:
    def __init__(self, use_gpu: bool = True, verbose: bool = False, summary_length: float = 0.0):
        """Returns object of InLegalNER class.
         It is used for loading Extractive Summarizer model in memory.
        Args:
            use_gpu (bool): Functionality to give a choice whether to use GPU for inference or not
             Setting it True doesn't ensure GPU will be utilized it need proper torch installation
            verbose (bool): When set to True will print info msg while inference
            summary_length (float): valid range(0-1) Length of summary to get in output. set it to 0 to use adaptive selection
        """
        self.__verbose__ = verbose
        if use_gpu:
            if torch.cuda.is_available():
                self.device = torch.device('cuda')
                msg.info('Extractive Summarizer will use GPU!')
            else:
                self.device = torch.device('cpu')
                msg.info('Extractive Summarizer will use CPU!')
        else:
            self.device = torch.device('cpu')
            msg.info('Extractive Summarizer will use CPU!')

        # load summarizer checkpoint
        state_dict = load_model_from_cache('ExtractiveSummarizer')

        # setup tokenizer
        self.bert_tokenizer = BertTokenizer.from_pretrained('bert-base-uncased', do_lower_case=True)
        self.__tokenizer__ = English().tokenizer

        # setup model arguments
        self.model_args, self.preprocessing_args = __setargs__()

        # setup model
        self.model = ExtSummarizer(self.model_args, self.device, state_dict)
        self.model.eval()
        if summary_length < 0 or summary_length > 1:
            summary_length_percentage = 0.0
            msg.info('Invalid input: Summary length need to be in range of 0-1. Setting it to adaptive')
        else:
            summary_length_percentage = float(summary_length)
        self.__summary_length_percentage__ = summary_length_percentage

    def _preprocess(self, input_data):
        try:
            if not input_data['annotations']:
                raise TypeError("Missing data in input for processing")
        except:
            raise TypeError("Invalid data format")
        bert_formatted = preprocess_for_summarization(input_data, self.bert_tokenizer, self.__tokenizer__)
        bert_formatted_and_tokenized = format_to_bert(bert_formatted, self.preprocessing_args)
        return bert_formatted_and_tokenized

    @staticmethod
    def _load_dataset(preprocessed_data):
        yield preprocessed_data

    def _inference(self, data_iter):
        def _get_ngrams(n, text):
            ngram_set = set()
            text_length = len(text)
            max_index_ngram_start = text_length - n
            for i in range(max_index_ngram_start + 1):
                ngram_set.add(tuple(text[i:i + n]))
            return ngram_set

        def _block_tri(c, p):
            tri_c = _get_ngrams(3, c.split())
            for s in p:
                tri_s = _get_ngrams(3, s.split())
                if len(tri_c.intersection(tri_s)) > 0:
                    return True
            return False

        self.model.eval()
        file_chunk_sent_scores = {}  ## key is filename and value is list of sentences containing sentence scores

        with torch.no_grad():
            for batch in data_iter:
                src = batch.src
                labels = batch.src_sent_labels
                segs = batch.segs
                clss = batch.clss
                mask = batch.mask_src
                mask_cls = batch.mask_cls
                sentence_rhetorical_roles = batch.sentence_rhetorical_roles
                sent_scores, mask = self.model(src, segs, clss, mask, mask_cls, sentence_rhetorical_roles)
                sent_scores = sent_scores.cpu().data.numpy()
                file_name, chunk_id, sentence_ids = batch.unique_ids[0].split('___')
                chunk_id = int(chunk_id)
                sentence_ids = ast.literal_eval(sentence_ids)
                src_labels = list(labels.cpu().numpy()[0])
                if type(sent_scores[0]) == np.float32:
                    sent_scores = np.array([sent_scores])
                sent_scores_list = list(sent_scores[0])
                sent_rhetorical_roles_list = list(sentence_rhetorical_roles.cpu().data.numpy()[0])
                for sent_id, (sent_txt, sent_label, sent_score, sent_rhet_role, sentence_id) in enumerate(
                        zip(batch.src_str[0], src_labels, sent_scores_list, sent_rhetorical_roles_list, sentence_ids)):
                    if file_chunk_sent_scores.get(file_name) is None:
                        file_chunk_sent_scores[file_name] = []
                    sent_dict = {'file_name': file_name, 'chunk_id': chunk_id, 'sent_id': sentence_id,
                                 'sent_txt': sent_txt,
                                 'sent_score': sent_score, 'sent_label': sent_label,
                                 'sent_rhetorical_role': sent_rhet_role}
                    file_chunk_sent_scores[file_name].append(sent_dict)

        return file_chunk_sent_scores

    def __call__(self, input_data):
        if self.__verbose__:
            msg.info('Processing documents with extractive summarizer model!!!')
        result = []
        for data in tqdm(input_data, disable=not self.__verbose__):
            task_id = data['id'].split('_')[-1]
            preprocessed_data = self._preprocess(data)
            data_iter = data_loader.Dataloader(self.model_args, self._load_dataset(preprocessed_data),
                                               self.model_args.test_batch_size, self.device,
                                               shuffle=False, is_test=True)
            inference_output = self._inference(data_iter)
            summary_texts = _postprocess(inference_output, data['annotations'], self.__summary_length_percentage__)
            result.append(copy.deepcopy({"id": 'ExtractiveSummarizer_' + task_id, "summaries": summary_texts}))
        return result



import warnings

import spacy

from opennyai.utils.sentencizer import process_nlp_in_chunks


def extract_entities_from_judgment_text(to_process, legal_nlp, mini_batch_size, do_sentence_level=True):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        judgement = to_process['judgement_doc']
        preamble = to_process['preamble_doc']
        if do_sentence_level:
            doc_judgement = get_sentence_docs(judgement, legal_nlp)
            doc_preamble = process_nlp_in_chunks(preamble.text, mini_batch_size, legal_nlp)
        else:
            doc_judgement = process_nlp_in_chunks(judgement.text, mini_batch_size, legal_nlp)
            doc_preamble = process_nlp_in_chunks(preamble.text, mini_batch_size, legal_nlp)

        ######### Combine preamble doc & judgement doc
        combined_doc = spacy.tokens.Doc.from_docs([doc_preamble, doc_judgement])

    return combined_doc


def get_sentence_docs(doc_judgment, nlp_judgment):
    sentences = [sent.text for sent in doc_judgment.sents]
    docs = []
    for doc in nlp_judgment.pipe(sentences):
        docs.append(doc)
    combined_docs = spacy.tokens.Doc.from_docs(docs)
    return combined_docs



"""
Conditional random field with weighting based on Lannoy et al. (2019) approach
"""
from typing import List, Tuple

import torch

from opennyai.rhetorical_roles.allennlp_helper.common.checks import ConfigurationError
from opennyai.rhetorical_roles.allennlp_helper.modules.conditional_random_field.conditional_random_field import (
    ConditionalRandomField,
)


class ConditionalRandomFieldWeightLannoy(ConditionalRandomField):
    """
    This module uses the "forward-backward" algorithm to compute
    the log-likelihood of its inputs assuming a conditional random field model.

    See, e.g. http://www.cs.columbia.edu/~mcollins/fb.pdf

    This is a weighted version of `ConditionalRandomField` which accepts a `label_weights`
    parameter to be used in the loss function in order to give different weights for each
    token depending on its label. The method implemented here is based on the paper
    *Weighted conditional random fields for supervised interpatient heartbeat
    classification* proposed by De Lannoy et. al (2019).
    See https://perso.uclouvain.be/michel.verleysen/papers/ieeetbe12gdl.pdf for more details.

    There are two other sample weighting methods implemented. You can find more details
    about them in: https://eraldoluis.github.io/2022/05/10/weighted-crf.html

    # Parameters

    num_tags : `int`, required
        The number of tags.
    label_weights : `List[float]`, required
        A list of weights to be used in the loss function in order to
        give different weights for each token depending on its label.
        `len(label_weights)` must be equal to `num_tags`. This is useful to
        deal with highly unbalanced datasets. The method implemented here was based on
        the paper *Weighted conditional random fields for supervised interpatient heartbeat
        classification* proposed by De Lannoy et. al (2019).
        See https://perso.uclouvain.be/michel.verleysen/papers/ieeetbe12gdl.pdf
    constraints : `List[Tuple[int, int]]`, optional (default = `None`)
        An optional list of allowed transitions (from_tag_id, to_tag_id).
        These are applied to `viterbi_tags()` but do not affect `forward()`.
        These should be derived from `allowed_transitions` so that the
        start and end transitions are handled correctly for your tag type.
    include_start_end_transitions : `bool`, optional (default = `True`)
        Whether to include the start and end transition parameters.
    """

    def __init__(
            self,
            num_tags: int,
            label_weights: List[float],
            constraints: List[Tuple[int, int]] = None,
            include_start_end_transitions: bool = True,
    ) -> None:
        super().__init__(num_tags, constraints, include_start_end_transitions)

        if label_weights is None:
            raise ConfigurationError("label_weights must be given")

        self.register_buffer("label_weights", torch.Tensor(label_weights))

    def forward(
            self, inputs: torch.Tensor, tags: torch.Tensor, mask: torch.BoolTensor = None
    ) -> torch.Tensor:
        """Computes the log likelihood for the given batch of input sequences $(x,y)$

        Args:
            inputs (torch.Tensor): (batch_size, sequence_length, num_tags) tensor of logits for the inputs $x$
            tags (torch.Tensor): (batch_size, sequence_length) tensor of tags $y$
            mask (torch.BoolTensor, optional): (batch_size, sequence_length) tensor of masking flags.
                Defaults to None.

        Returns:
            torch.Tensor: (batch_size,) log likelihoods $log P(y|x)$ for each input
        """
        if mask is None:
            mask = torch.ones(*tags.size(), dtype=torch.bool, device=inputs.device)
        else:
            # The code below fails in weird ways if this isn't a bool tensor, so we make sure.
            mask = mask.to(torch.bool)

        log_denominator = self._input_likelihood_lannoy(inputs, tags, mask)
        log_numerator = self._joint_likelihood_lannoy(inputs, tags, mask)

        return torch.sum(log_numerator - log_denominator)

    def _input_likelihood_lannoy(
            self, logits: torch.Tensor, tags: torch.Tensor, mask: torch.BoolTensor
    ) -> torch.Tensor:
        """
        Computes the (batch_size,) denominator term for the log-likelihood, which is the
        sum of the likelihoods across all possible state sequences.

        Compute this value using the scaling trick instead of the log domain trick, since
        this is necessary to implement the label-weighting method by Lannoy et al. (2012).
        """
        batch_size, sequence_length, num_tags = logits.size()

        # Transpose batch size and sequence dimensions
        mask = mask.transpose(0, 1).contiguous()
        logits = logits.transpose(0, 1).contiguous()
        tags = tags.transpose(0, 1).contiguous()

        # insert an 1-sized second dimension to match z.shape
        label_weights = self.label_weights.view(num_tags, 1)

        # emit_scores.shape = (batch_size, num_tags)
        emit_scores = logits[0]

        # Initial alpha is the (batch_size, num_tags) tensor of likelihoods combining the
        # transitions to the initial states and the logits for the first timestep.
        # alpha.shape = (batch_size, num_tags)
        if self.include_start_end_transitions:
            alpha = torch.exp(self.start_transitions.view(1, num_tags) + emit_scores)
        else:
            alpha = torch.exp(emit_scores)

        # z.shape = (batch_size, 1)
        z = alpha.sum(dim=1, keepdim=True)
        alpha = alpha / z
        sum_log_z = torch.log(z) * label_weights[tags[0]]

        # For each i we compute logits for the transitions from timestep i-1 to timestep i.
        # We do so in a (batch_size, num_tags, num_tags) tensor where the axes are
        # (instance, current_tag, next_tag)
        for i in range(1, sequence_length):
            # multiply the logits by the label weights
            # logits[i].shape: (batch_size, num_tags)
            # emit_scores = torch.mul(logits[i], label_weights)
            emit_scores = logits[i]

            # The emit scores are for time i ("next_tag") so we broadcast along the current_tag axis.
            emit_scores = emit_scores.view(batch_size, 1, num_tags)
            # Transition scores are (current_tag, next_tag) so we broadcast along the instance axis.
            transition_scores = self.transitions.view(1, num_tags, num_tags)
            # Alpha is for the current_tag (i-1), so we broadcast along the next_tag axis.
            broadcast_alpha = alpha.view(batch_size, num_tags, 1)

            # Add all the scores together and logexp over the current_tag axis.
            inner = broadcast_alpha * torch.exp(emit_scores + transition_scores)

            # In valid positions (mask == True) we want to take the logsumexp over the current_tag dimension
            # of `inner`. Otherwise (mask == False) we want to retain the previous alpha.
            alpha = inner.sum(dim=1) * mask[i].view(batch_size, 1) + alpha * (~mask[i]).view(
                batch_size, 1
            )

            # scale alphas to avoid underflow (sum of alphas equal to 1)
            z = alpha.sum(dim=1, keepdim=True)
            alpha = alpha / z
            # weight z (normalization factor) according to the current tag
            sum_log_z += torch.log(z) * label_weights[tags[i]]

        # Every sequence needs to end with a transition to the stop_tag.
        if self.include_start_end_transitions:
            alpha = alpha * torch.exp(self.end_transitions.view(1, num_tags))
            z = alpha.sum(dim=1, keepdim=True)
            # alpha = alpha / z  # this step is unnecessary since alpha is not used anymore
            sum_log_z += torch.log(z)

        return sum_log_z.squeeze(1)

    def _joint_likelihood_lannoy(
            self, logits: torch.Tensor, tags: torch.Tensor, mask: torch.BoolTensor
    ) -> torch.Tensor:
        """
        Computes the numerator term for the log-likelihood, which is just score(inputs, tags)
        """
        batch_size, sequence_length, _ = logits.data.shape

        # Transpose batch size and sequence dimensions:
        logits = logits.transpose(0, 1).contiguous()
        mask = mask.transpose(0, 1).contiguous()
        tags = tags.transpose(0, 1).contiguous()

        # Start with the transition scores from start_tag to the first tag in each input
        if self.include_start_end_transitions:
            score = self.start_transitions.index_select(0, tags[0])
        else:
            score = 0.0

        label_weights = self.label_weights

        # weight transition score using current_tag, i.e., t(i,j) will be t(i,j)*w(i),
        # where t(i,j) is the score to transition from i to j and w(i) is the weight
        # for tag i.
        transitions = self.transitions * label_weights.view(-1, 1)

        # Add up the scores for the observed transitions and all the inputs but the last
        for i in range(sequence_length - 1):
            # Each is shape (batch_size,)
            current_tag, next_tag = tags[i], tags[i + 1]

            # The scores for transitioning from current_tag to next_tag
            transition_score = transitions[current_tag.view(-1), next_tag.view(-1)]

            # The score for using current_tag
            emit_score = logits[i].gather(1, current_tag.view(batch_size, 1)).squeeze(1)

            # Weight emit scores by label.
            emit_score *= label_weights[current_tag.view(-1)]

            # Include transition score if next element is unmasked,
            # input_score if this element is unmasked.
            score = score + transition_score * mask[i + 1] + emit_score * mask[i]

        # Transition from last state to "stop" state. To start with, we need to find the last tag
        # for each instance.
        last_tag_index = mask.sum(0).long() - 1
        last_tags = tags.gather(0, last_tag_index.view(1, batch_size)).squeeze(0)

        # Compute score of transitioning to `stop_tag` from each "last tag".
        if self.include_start_end_transitions:
            last_transition_score = self.end_transitions.index_select(0, last_tags)
        else:
            last_transition_score = 0.0

        # Add the last input if it's not masked.
        last_inputs = logits[-1]  # (batch_size, num_tags)
        last_input_score = last_inputs.gather(1, last_tags.view(-1, 1))  # (batch_size, 1)
        last_input_score = last_input_score.squeeze()  # (batch_size,)

        # Weight last emit scores by label weights.
        last_input_score = last_input_score * label_weights[last_tags.view(-1)]

        score = score + last_transition_score + last_input_score * mask[-1]

        return score



import copy

from wasabi import msg

import opennyai.ner as InLegalNER
from opennyai.rhetorical_roles.rhetorical_roles import RhetoricalRolePredictor
from opennyai.summarizer.ExtractiveSummarizer import ExtractiveSummarizer


class Pipeline:
    def __init__(self, components=None, use_gpu=True, verbose=True, ner_model_name='en_legal_ner_trf',
                 ner_mini_batch_size=40000, ner_do_sentence_level=True, ner_do_postprocess=True,
                 ner_statute_shortforms_path='', summarizer_summary_length=0.0):
        self.__ner_mini_batch_size__ = ner_mini_batch_size
        self.__do_sentence_level__ = ner_do_sentence_level
        self.__do_postprocess__ = ner_do_postprocess
        self.__statute_shortforms_path__ = ner_statute_shortforms_path
        default_component_values = ['NER', 'Rhetorical_Role', 'Summarizer']
        if components is None:
            components = default_component_values
        if 'Summarizer' in components:
            if 'Rhetorical_Role' not in components:
                components.append('Rhetorical_Role')
        assert all(
            [i in default_component_values for i in components]), f"Invalid component value given in {components}"
        self.components = components
        self.__verbose__ = verbose
        if 'NER' in components:
            if self.__verbose__:
                msg.info('Loading NER...')
            self.__ner_extractor__ = InLegalNER.load(use_gpu=use_gpu, model_name=ner_model_name)

        if 'Rhetorical_Role' in components or 'Summarizer' in components:
            if self.__verbose__:
                msg.info('Loading Rhetorical Role...')
            self.__rr_model__ = RhetoricalRolePredictor(use_gpu=use_gpu, verbose=verbose)

        if 'Summarizer' in components:
            if self.__verbose__:
                msg.info('Loading Extractive summarizer...')
            self.__summarizer__ = ExtractiveSummarizer(use_gpu=use_gpu, verbose=verbose,
                                                       summary_length=summarizer_summary_length)

    @staticmethod
    def __combine_model_outputs__(ner_json_results=None, rr_output=None, summary_output=None):
        '''combines the outputs of 3 models into single list'''
        combined_results = {}
        ####### Add NER results
        if ner_json_results:
            for doc_ner in ner_json_results:
                doc_id = doc_ner['id'].split('_')[1]
                if combined_results.get(doc_id) is None:
                    combined_results[doc_id] = {'id': doc_id, 'data': {'text': doc_ner['data']['text'],
                                                                       'preamble_end_char_offset': doc_ner['data'].get(
                                                                           'preamble_end_char_offset')}}

                combined_results[doc_id]['annotations'] = doc_ner['annotations']

        ####### Add RR results
        if rr_output:
            for doc_rr in rr_output:
                doc_id = doc_rr['id'].split('_')[1]
                if combined_results.get(doc_id) is None:
                    combined_results[doc_id] = {'id': doc_id,
                                                'data': {'text': doc_rr['data']['text'],
                                                         'preamble_end_char_offset': doc_rr['data'].get(
                                                             'preamble_end_char_offset')}}

                combined_results[doc_id]['annotations'] = doc_rr['annotations']

        ####### Add summary results
        if summary_output:
            for doc_summary in summary_output:
                doc_id = doc_summary['id'].split('_')[1]
                combined_results[doc_id]['summary'] = doc_summary['summaries']

        if ner_json_results and rr_output:
            for doc_ner in ner_json_results:
                entities = [entity for annotation in doc_ner['annotations'] for entity in
                            annotation['entities']]
                doc_id = doc_ner['id'].split('_')[1]
                for entity in entities:
                    for sent in combined_results[doc_id]['annotations']:
                        if sent.get('entities') is None:
                            sent['entities'] = []
                        if entity['start'] >= sent['start'] and entity['end'] <= sent['end']:
                            sent['entities'].append(entity)
                            break
                combined_results[doc_id]['annotations'] = copy.deepcopy(combined_results[doc_id]['annotations'])

        return [result for doc_id, result in combined_results.items()]

    @staticmethod
    def __postprocess_ner_to_sentence_level__(doc):
        id = "LegalNER_" + doc.user_data['doc_id']
        final_output = InLegalNER.get_json_from_spacy_doc(doc)
        output = {'id': id, 'annotations': [],
                  'data': {'text': doc.text, 'original_text': doc.user_data['original_text'],
                           'preamble_end_char_offset': doc.user_data['preamble_end_char_offset']}}
        for sent in doc.sents:
            import uuid
            uid = uuid.uuid4()
            id = uid.hex
            temp = copy.deepcopy({"id": id,
                                  "start": sent.start_char,
                                  "end": sent.end_char,
                                  "text": sent.text,
                                  "entities": []})
            for entity in final_output['annotations']:
                if entity['start'] >= temp['start'] and entity['end'] <= temp['end']:
                    temp['entities'].append(entity)
            output['annotations'].append(temp)
        return output

    def __call__(self, data):
        ner_json_results, self._ner_model_output, self._rr_model_output, self._summarizer_model_output = None, None, None, None
        if 'NER' in self.components:
            self._ner_model_output = self.__ner_extractor__(data, verbose=self.__verbose__,
                                                            do_sentence_level=self.__do_sentence_level__,
                                                            do_postprocess=self.__do_postprocess__,
                                                            mini_batch_size=self.__ner_mini_batch_size__,
                                                            statute_shortforms_path=self.__statute_shortforms_path__)
            if not isinstance(self._ner_model_output, list):
                self._ner_model_output = [self._ner_model_output]
            ner_json_results = [self.__postprocess_ner_to_sentence_level__(doc) for doc in self._ner_model_output]

        if 'Rhetorical_Role' in self.components or 'Summarizer' in self.components:
            self._rr_model_output = self.__rr_model__(data)
        if 'Summarizer' in self.components:
            self._summarizer_model_output = self.__summarizer__(self._rr_model_output)

        return self.__combine_model_outputs__(ner_json_results, self._rr_model_output, self._summarizer_model_output)



import os

from opennyai.utils.download import CACHE_DIR

EXTRACTIVE_SUMMARIZER_CACHE_PATH = os.path.join(CACHE_DIR, 'ExtractiveSummarizer'.lower())


class ARGS:
    def __init__(self):
        self.default = True


def __setargs__():
    parser = ARGS()
    parser.task = 'ext'
    parser.encoder = 'bert'
    parser.mode = 'test'
    parser.bert_data_path = EXTRACTIVE_SUMMARIZER_CACHE_PATH
    parser.model_path = EXTRACTIVE_SUMMARIZER_CACHE_PATH
    parser.result_path = EXTRACTIVE_SUMMARIZER_CACHE_PATH
    parser.temp_dir = EXTRACTIVE_SUMMARIZER_CACHE_PATH
    parser.batch_size = 5000
    parser.test_batch_size = 1
    parser.max_pos = 512
    parser.use_interval = True
    parser.large = False
    parser.load_from_extractive = ''
    parser.sep_optim = True
    parser.lr_bert = 2e-3
    parser.lr_dec = 2e-3
    parser.use_bert_emb = False
    parser.share_emb = False
    parser.finetune_bert = True
    parser.dec_dropout = 0.2
    parser.dec_layers = 6
    parser.dec_hidden_size = 768
    parser.dec_heads = 8
    parser.dec_ff_size = 2048
    parser.enc_hidden_size = 512
    parser.enc_ff_size = 512
    parser.enc_dropout = 0.2
    parser.enc_layers = 6
    # params for EXT
    parser.ext_dropout = 0.2
    parser.ext_layers = 2
    parser.ext_hidden_size = 768
    parser.ext_heads = 8
    parser.ext_ff_size = 2048
    parser.label_smoothing = 0.1
    parser.generator_shard_size = 32
    parser.alpha = 0.95
    parser.beam_size = 5
    parser.min_length = 0
    parser.max_length = 2000
    parser.max_tgt_len = 140
    parser.use_rhetorical_roles = True  #### whether to use rhetorical roles in the mode
    parser.seperate_summary_for_each_rr = False  #### whether to select top N sentences from each rhetorical role
    parser.rogue_exclude_roles_not_in_test = True  #### whether to remove the sections that are present in predicted summaries which are not in test data while ROGUE calculation
    parser.add_additional_mandatory_roles_to_summary = False  #### whether to add the additional mandatory roles to predicted summary
    parser.summary_sent_precent = 20  ##### top N percentage of sentences to be selected
    parser.use_adaptive_summary_sent_percent = True  ##### whether summary sentence percentage should be chosen as per input text sentence length
    parser.param_init = 0
    parser.param_init_glorot = True
    parser.optim = 'adam'
    parser.lr = 1
    parser.beta1 = 0.9
    parser.beta2 = 0.999
    parser.warmup_steps = 8000
    parser.warmup_steps_bert = 8000
    parser.warmup_steps_dec = 8000
    parser.max_grad_norm = 0
    parser.save_checkpoint_steps = 5
    parser.accum_count = 1
    parser.report_every = 1
    parser.train_steps = 1000
    parser.recall_eval = False
    parser.visible_gpus = '-1'
    parser.gpu_ranks = '0'
    parser.log_file = './log.log'
    parser.seed = 666
    parser.test_all = False
    parser.test_from = ''
    parser.test_start_from = -1
    parser.train_from = ''
    parser.report_rouge = True
    parser.block_trigram = True
    model_args = parser

    parser_preprocessing = ARGS()
    parser_preprocessing.pretrained_model = 'bert'

    parser_preprocessing.mode = 'format_to_bert'
    parser_preprocessing.select_mode = 'greedy'
    parser_preprocessing.map_path = EXTRACTIVE_SUMMARIZER_CACHE_PATH
    parser_preprocessing.raw_path = EXTRACTIVE_SUMMARIZER_CACHE_PATH
    parser_preprocessing.save_patt = EXTRACTIVE_SUMMARIZER_CACHE_PATH

    parser_preprocessing.shard_size = 2000
    parser_preprocessing.min_src_nsents = 0
    parser_preprocessing.max_src_nsents = 50000
    parser_preprocessing.min_src_ntokens_per_sent = 0
    parser_preprocessing.max_src_ntokens_per_sent = 512
    parser_preprocessing.min_tgt_ntokens = 0
    parser_preprocessing.max_tgt_ntokens = 20000

    parser_preprocessing.lower = True
    parser_preprocessing.use_bert_basic_tokenizer = False

    parser_preprocessing.log_file = os.path.join(EXTRACTIVE_SUMMARIZER_CACHE_PATH, 'log.log')

    parser_preprocessing.dataset = ''

    parser_preprocessing.n_cpus = 2
    return model_args, parser_preprocessing



import os

import numpy as np
from sklearn.model_selection import KFold, train_test_split

from .batch_creator import BatchCreator
from .dataset_reader import DocumentsDataset

PUBMED_LABELS = ["DEFAULT", 'mask', "NONE", "PREAMBLE", "FAC", "ISSUE", "ARG_RESPONDENT", "ARG_PETITIONER", "ANALYSIS",
                 "PRE_RELIED", "PRE_NOT_RELIED", "STA", "RLC", "RPC", "RATIO"]
PUBMED_LABELS_PRES = ["DEFAULT", 'mask', "NONE", "PREAMBLE", "FAC", "ISSUE", "ARG_RESPONDENT", "ARG_PETITIONER",
                      "ANALYSIS", "PRE_RELIED", "PRE_NOT_RELIED", "STA", "RLC", "RPC", "RATIO"]

GEN_LABELS = ["PREAMBLE", "NONE", "FAC", "ISSUE", "ARG_RESPONDENT", "ARG_PETITIONER", "ANALYSIS", "PRE_RELIED",
              "PRE_NOT_RELIED", "STA", "RLC", "RPC", "RATIO"]
GEN_LABELS_PRES = GEN_LABELS

DRI_TASK = "DRI"
PUBMED_TASK = "pubmed-20k"
NICTA_TASK = "nicta_piboso"
ART_TASK = "ART"

GEN_DRI_TASK = "DRI_generic"
GEN_PMD_TASK = "PMD_generic"
GEN_NIC_TASK = "NIC_generic"
GEN_ART_TASK = "ART_generic"


def tgeneric_task(task_name, train_batch_size, max_docs):
    return Task(task_name, GEN_LABELS,
                train_batch_size, 1, max_docs, short_name=task_name,
                labels_pres=GEN_LABELS_PRES)


def pubmed_task(train_batch_size, max_docs, verbose, data_folder="datasets/"):
    return Task(PUBMED_TASK, PUBMED_LABELS,
                train_batch_size, 1, max_docs, short_name="PMD", labels_pres=PUBMED_LABELS_PRES,
                data_folder=data_folder, verbose=verbose)


def pubmed_task_small(train_batch_size, max_docs):
    return Task(PUBMED_TASK + "_small", PUBMED_LABELS,
                train_batch_size, 1, max_docs,
                portion_training_data=1.0 / 20.0, task_folder_name=PUBMED_TASK, short_name="mPMD",
                labels_pres=PUBMED_LABELS_PRES)


class Fold:
    def __init__(self, train, dev, test):
        self.train = train
        self.dev = dev
        self.test = test


class Task:
    def __init__(self, task_name, labels, train_batch_size, num_fols, max_docs=-1,
                 dev_metric="weighted-f1", portion_training_data=1.0,
                 task_folder_name=None, short_name=None,
                 labels_pres=None,
                 data_folder="datasets/",
                 verbose=False):
        self.labels_pres = labels_pres
        self.short_name = short_name
        self.task_name = task_name
        self.labels = labels
        # self.data_dir = os.path.join(data_folder, task_name if task_folder_name is None else task_folder_name)
        self.data_dir = data_folder
        self.train_batch_size = train_batch_size
        self.num_folds = num_fols
        self.max_docs = max_docs
        self.folds = None
        self.folds_examples = None
        self.dev_metric = dev_metric
        self.portion_training_data = portion_training_data
        self.__verbose__ = verbose

    def get_labels_pres_titled(self):
        '''Labels ordered in presentation-order titled. '''
        return [l.title() for l in self.labels_pres]

    def get_labels_titled(self):
        '''Labels titled. '''
        return [l.title() for l in self.labels]

    def _get_batches(self, examples):
        ds_builder = BatchCreator(
            examples,
            tokenizer=None,
            labels=self.labels,
            batch_sentence_size=self.train_batch_size,
            max_seq_length=None
        )
        batches = ds_builder.get_batches(task_name=self.task_name)
        return batches

    def _load_full_set(self, file_suffix='scibert'):
        '''Returns one Fold. '''
        # if self.__verbose__:
        #     msg.info("Loading tokenized data...")
        full_examples = DocumentsDataset(os.path.join(self.data_dir, f"full_{file_suffix}.txt"), max_docs=self.max_docs)
        # if self.__verbose__:
        #     msg.info("Loading tokenized data finished.")
        return list(full_examples)

    def _load_train_dev_test_examples(self, file_suffix='scibert') -> Fold:
        # if self.__verbose__:
        #     msg.info("Loading tokenized data...")

        train_examples = DocumentsDataset(os.path.join(self.data_dir, f"train_{file_suffix}.txt"),
                                          max_docs=self.max_docs)
        dev_examples = DocumentsDataset(os.path.join(self.data_dir, f"dev_{file_suffix}.txt"), max_docs=self.max_docs)
        test_examples = DocumentsDataset(os.path.join(self.data_dir, f"test_{file_suffix}.txt"), max_docs=self.max_docs)

        train_examples = self.truncate_train_examples(train_examples)

        # if self.__verbose__:
        #     msg.info("Loading tokenized data finished.")
        return [(train_examples, dev_examples, test_examples)]

    def truncate_train_examples(self, train_examples):
        if self.portion_training_data < 1.0:
            train_examples = list(train_examples)
            new_len = int(len(train_examples) * self.portion_training_data)
            # if self.__verbose__:
            #     msg.info(
            #         f"Truncating training examples with factor {self.portion_training_data} from {len(train_examples)} to {new_len}")
            train_examples = train_examples[0: new_len]
        return train_examples

    def get_all_examples(self, file_suffix='scibert'):
        if self.num_folds == 1:
            train, dev, test = self._load_train_dev_test_examples(file_suffix)[0]
            all_examples = []
            all_examples += list(train)
            print(len(all_examples))
            all_examples += list(dev)
            print(len(all_examples))
            all_examples += list(test)
            print(len(all_examples))
            return all_examples
        else:
            return self._load_full_set(file_suffix)

    def get_folds_examples(self, file_suffix='scibert'):
        if self.folds_examples is not None:
            return self.folds_examples
        self.folds_examples = []
        # if self.__verbose__:
        #     msg.info(f"Loading data with {self.num_folds} folds...")
        if self.num_folds == 1:
            self.folds_examples = self._load_train_dev_test_examples(file_suffix=file_suffix)
        else:
            full_examples = np.array(self._load_full_set(file_suffix=file_suffix))
            # assert len(full_examples) % self.num_folds == 0

            self.folds = []
            kf = KFold(n_splits=self.num_folds, shuffle=True, random_state=268)
            for train_index, test_index in kf.split(full_examples):
                train_and_dev = full_examples[train_index]
                test = full_examples[test_index]
                train, dev = train_test_split(train_and_dev, test_size=1.0 / self.num_folds, shuffle=False)

                train = self.truncate_train_examples(train)

                self.folds_examples.append((train, dev, test))

        return self.folds_examples

    def get_folds(self):
        if self.folds is not None:
            return self.folds

        folds_examples = self.get_folds_examples()
        self.folds = []
        # if self.__verbose__:
        #     msg.info(f"Creating batches for {self.num_folds} folds...")
        for train, dev, test in folds_examples:
            train_batches = self._get_batches(train)
            dev_batches = self._get_batches(dev)
            test_batches = self._get_batches(test)

            self.folds.append(Fold(train_batches, dev_batches, test_batches))
        # if self.__verbose__:
        #     msg.info(f"Creating batches finished.")
        return self.folds

    def get_stats_counts(self):
        counts = dict()
        all_examples = self.get_all_examples()
        counts["docs"] = len(all_examples)
        counts["sentences"] = 0
        for d in all_examples:
            for l in d.labels:
                counts["sentences"] += 1
                if l in counts:
                    counts[l] += 1
                else:
                    counts[l] = 1

        return counts

    def get_test_label_counts(self, fold_num):
        fold_num = fold_num % self.num_folds
        _, _, test = self.get_folds_examples()[fold_num]

        counts = [0] * len(self.labels)
        for d in test:
            for l in d.labels:
                label_id = self.labels.index(l)
                counts[label_id] += 1
        return counts




from opennyai import Pipeline
from opennyai.utils import Data
from spacy import displacy
from opennyai.ner.ner_utils import ner_displacy_option


#text = open('/Users/prathamesh/Desktop/itat_sample2.txt').read()
#text = open('/Users/prathamesh/tw_projects/OpenNyAI/git_repos/Opennyai/tests/examples/judgement_texts/72703592.txt').read()
text = open('/Users/prathamesh/tw_projects/OpenNyAI/git_repos/Opennyai/tests/examples/judgement_texts/811682.txt').read()

data = Data([text])
pipeline = Pipeline(components=['NER'], use_gpu=False, verbose=True)
results = pipeline(data)
ner_doc = pipeline._ner_model_output[0]


displacy.serve(ner_doc, style='ent',port=8083
               ,options=ner_displacy_option)
# from opennyai import Pipeline
# from opennyai.utils import Data
# import glob
# import json
# import pandas as pd
# import re
# from tqdm.notebook import tqdm
# import spacy
#
# search_results = pd.read_csv(
#     '/Users/prathamesh/tw_projects/OpenNyAI/data/court_search/personal_liberty/search_results_judgment_metadata.csv')
# search_result_dir = '/Users/prathamesh/tw_projects/OpenNyAI/data/court_search/personal_liberty/'
# search_results['judgment_text'] = search_results['judgment_id'].apply(
#     lambda x: open(search_result_dir + 'txt/' + x + '.txt').read())
# files = search_results['judgment_text'].to_list()
# ids = search_results['judgment_id'].to_list()
#
#
# files = []
# ids = []
# nlp = spacy.load('en_core_web_sm')
# for index, row in search_results.iterrows():
#     if len(nlp.tokenizer(row['judgment_text'])) < 100000:
#         files.append(row['judgment_text'])
#         ids.append(row['judgment_id'])
#
# data = Data(input_text=files[:2],file_ids= ids[:2], mini_batch_size=20000)
# data = [i for i in data]
# for i in data:
#     print(i['file_id'])


import re
from urllib.request import urlopen, Request

from bs4 import BeautifulSoup as soup, Tag


def get_useful_text_from_indiankanoon_html_tag(ik_tag):
    tag_txt = ''
    for content in ik_tag.contents:
        if isinstance(content, Tag):
            if not (content.get('class') is not None and 'hidden_text' in content['class']):
                tag_txt = tag_txt + content.text
        else:
            tag_txt = tag_txt + str(content)
    return tag_txt


def check_hidden_text_is_invalid(text):
    return True  ## Most of the times hiddent text is garbage
    # if not bool(re.match('[a-zA-Z]',"".join(text.split()))):
    #     return True
    # elif bool(re.match('::: (Uploaded on - |Downloaded on -)+ .*?:::',text)):
    #     return True
    # else:
    #     return False


def get_text_from_indiankanoon_url(url: str) -> str:
    """Returns text from Indiakannon URL.
    """
    req = Request(url, headers={'User-Agent': 'Mozilla/5.0'})

    try:
        webpage = urlopen(req, timeout=10).read()
        page_soup = soup(webpage, "html.parser")

        judgment_txt_tags = page_soup.find_all(['p', 'blockquote', 'pre'])
        judgment_txt = ''
        for judgment_txt_tag in judgment_txt_tags:
            tag_txt = ''
            if judgment_txt_tag.get('id') is not None and (judgment_txt_tag['id'].startswith('p_') or
                                                           judgment_txt_tag['id'].startswith('blockquote_') or
                                                           judgment_txt_tag['id'].startswith('pre_')):
                for content in judgment_txt_tag.contents:
                    if isinstance(content, Tag):
                        if content.get('class') is not None and 'hidden_text' in content['class']:
                            if not check_hidden_text_is_invalid(content.text.strip()):
                                tag_txt = tag_txt + str(content)
                        else:
                            tag_txt = tag_txt + content.text
                    else:
                        tag_txt = tag_txt + str(content)

                if not judgment_txt_tag['id'].startswith('pre_'):
                    ##### remove unwanted formating except for pre_ tags
                    tag_txt = re.sub(r'\s+(?!\s*$)', ' ',
                                     tag_txt)  ###### replace the multiple spaces, newlines with space except for the ones at the end.
                    tag_txt = re.sub(r'([.\"\?])\n', r'\1 \n\n',
                                     tag_txt)  ###### add the extra new line for correct sentence breaking in spacy
                    tag_txt = re.sub(r'\n{2,}', '\n\n', tag_txt)

                judgment_txt = judgment_txt + tag_txt

    except:
        judgment_txt = ''

        ###### remove known footer, header patterns
    regex_patterns_to_remove = [{'pattern': r'http://www.judis.nic.in(\s*?\x0c\s*?)?'},
                                {
                                    'pattern': r'(::: Uploaded on - \d\d/\d\d/\d\d\d\d\s+)?::: Downloaded on - .{5,50}:::'},
                                {'pattern': r'https://www.mhc.tn.gov.in/judis/(\s*?\x0c\s*?)?'},
                                {
                                    'pattern': r'Signature Not Verified Signed By:.{5,100}Signing Date:\d\d\.\d\d\.\d\d\d\d(.{1,50}Page \d+\s*?! of \d+\s*?!\s*?\d\d:\d\d:\d\d)?',
                                    'flags': re.DOTALL | re.IGNORECASE},
                                ]
    for pattern_dict in regex_patterns_to_remove:
        if pattern_dict.get('flags') is not None:
            judgment_txt = re.sub(pattern_dict['pattern'], "", judgment_txt, flags=pattern_dict['flags'])
        else:
            judgment_txt = re.sub(pattern_dict['pattern'], "", judgment_txt)

    return judgment_txt.strip()



import os
import shutil
import subprocess
import sys
from pathlib import Path

import torch

"""Functions for downloading opennyai ner models."""
PIP_INSTALLER_URLS = {
    "en_legal_ner_trf": "https://huggingface.co/opennyaiorg/en_legal_ner_trf/resolve/main/en_legal_ner_trf-any-py3-none-any.whl",
    "en_legal_ner_sm": "https://huggingface.co/opennyaiorg/en_legal_ner_sm/resolve/main/en_legal_ner_sm-any-py3-none-any.whl",
    "en_core_web_md": "https://github.com/explosion/spacy-models/releases/download/en_core_web_md-3.8.0/en_core_web_md-3.8.0-py3-none-any.whl",
    "en_core_web_sm": "https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-3.8.0/en_core_web_sm-3.8.0-py3-none-any.whl",
    "en_core_web_trf": "https://github.com/explosion/spacy-models/releases/download/en_core_web_trf-3.8.0/en_core_web_trf-3.8.0-py3-none-any.whl",
}
TORCH_PT_MODEL_URLS = {
    "RhetoricalRole": "https://storage.googleapis.com/indianlegalbert/OPEN_SOURCED_FILES/Rhetorical_Role_Benchmark/Model/model.pt",
    "ExtractiveSummarizer": "https://storage.googleapis.com/indianlegalbert/OPEN_SOURCED_FILES/Extractive_summarization/model/model_headnotes/model.pt"
}
CACHE_DIR = os.path.join(str(Path.home()), '.opennyai')


def install(package: str):
    """
    It is used for installing pip wheel file for model supported
    Args:
        package (string): wheel file url
    """
    import tempfile
    import urllib.request

    # Download wheel to temp dir, renaming if it has an invalid version
    filename = package.rsplit("/", 1)[-1]
    tmp_dir = tempfile.mkdtemp()
    local_path = os.path.join(tmp_dir, filename)
    urllib.request.urlretrieve(package, local_path)

    # Fix wheels with invalid version in filename by reading actual version from metadata
    if "-any-py3-" in filename:
        import zipfile
        with zipfile.ZipFile(local_path) as zf:
            for name in zf.namelist():
                if name.endswith("/METADATA"):
                    metadata = zf.read(name).decode()
                    for line in metadata.splitlines():
                        if line.startswith("Version:"):
                            version = line.split(":", 1)[1].strip()
                            fixed_filename = filename.replace("-any-py3-", f"-{version}-py3-")
                            fixed_path = os.path.join(tmp_dir, fixed_filename)
                            os.rename(local_path, fixed_path)
                            local_path = fixed_path
                            break
                    break

    uv_path = shutil.which("uv")
    env = os.environ.copy()
    try:
        if uv_path:
            subprocess.check_call(
                [uv_path, "pip", "install", local_path, "--no-deps"],
                stdout=subprocess.DEVNULL, env=env,
            )
        else:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", local_path, "--no-deps"],
                stdout=subprocess.DEVNULL,
            )
    finally:
        import shutil as _shutil
        _shutil.rmtree(tmp_dir, ignore_errors=True)


def load_model_from_cache(model_name: str):
    """
    It is used for downloading model.pt files supported and developed by Opennyai
    Args:
        model_name (string): model name to download and save
    """
    if TORCH_PT_MODEL_URLS.get(model_name) is None:
        raise RuntimeError(f'{model_name} is not supported by opennyai, please check the name!')
    else:
        model_url = TORCH_PT_MODEL_URLS[model_name]
        os.makedirs(os.path.join(CACHE_DIR, model_name.lower()), exist_ok=True)
        return torch.hub.load_state_dict_from_url(model_url, model_dir=os.path.join(CACHE_DIR, model_name.lower()),
                                                  check_hash=True, map_location=torch.device('cpu'))



import datetime
import sys
from threading import Lock

import torch
from prettytable import PrettyTable


def log(str):
    print(str, file=sys.stderr)
    sys.stderr.flush()


def get_device(gpu_device=0):
    device = torch.device(f"cuda:{gpu_device}" if torch.cuda.is_available() else "cpu")
    return device


def tensor_dict_to_gpu(d, device):
    tensor_dict_to_device(d, device)


def tensor_dict_to_cpu(d):
    tensor_dict_to_device(d, torch.device("cpu"))


def tensor_dict_to_device(d, device):
    for k, v in d.items():
        if torch.is_tensor(v):
            d[k] = v.to(device)
        else:
            d[k] = v


class ResultWriter:
    def __init__(self, results_filename):
        self.results_filename = results_filename
        self.lock = Lock()

    def write(self, str):
        self.lock.acquire()
        try:
            with open(self.results_filename, "a", encoding="utf-8") as f:
                f.write(str + "\n")
        finally:
            self.lock.release()

    def log(self, msg):
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        msg = timestamp + ": " + msg
        log(msg)
        self.lock.acquire()
        try:
            with open(self.results_filename + ".log", "a", encoding="utf-8") as f:
                f.write(msg + "\n")
        finally:
            self.lock.release()


def get_num_model_parameters(model):
    return sum([p.numel() for p in model.parameters() if p.requires_grad])


def print_model_parameters(model):
    table = PrettyTable(["Modules", "Parameters"])
    total_params = 0
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad: continue
        param = parameter.numel()
        table.add_row([name, f'{param:,}'])
        total_params += param
    print(table)
    print(f"Total Trainable Params: {total_params}")
    return total_params



"""
Modules that transform a sequence of input vectors
into a sequence of output vectors.
Some are just basic wrappers around existing PyTorch modules,
others are AllenNLP modules.

The available Seq2Seq encoders are

- `"gru"` : allennlp.modules.seq2seq_encoders.GruSeq2SeqEncoder
- `"lstm"` : allennlp.modules.seq2seq_encoders.LstmSeq2SeqEncoder
- `"rnn"` : allennlp.modules.seq2seq_encoders.RnnSeq2SeqEncoder
- `"augmented_lstm"` : allennlp.modules.seq2seq_encoders.AugmentedLstmSeq2SeqEncoder
- `"alternating_lstm"` : allennlp.modules.seq2seq_encoders.StackedAlternatingLstmSeq2SeqEncoder
- `"pass_through"` : allennlp.modules.seq2seq_encoders.PassThroughEncoder
- `"feedforward"` : allennlp.modules.seq2seq_encoders.FeedForwardEncoder
- `"pytorch_transformer"` : allennlp.modules.seq2seq_encoders.PytorchTransformer
- `"compose"` : allennlp.modules.seq2seq_encoders.ComposeEncoder
- `"gated-cnn-encoder"` : allennlp.momdules.seq2seq_encoders.GatedCnnEncoder
- `"stacked_bidirectional_lstm"`: allennlp.modules.seq2seq_encoders.StackedBidirectionalLstmSeq2SeqEncoder
"""



# from allennlp.modules.conditional_random_field.conditional_random_field import (
#     ConditionalRandomField,
#     allowed_transitions,
# )
# from allennlp.modules.conditional_random_field.conditional_random_field_wemission import (
#     ConditionalRandomFieldWeightEmission,
# )
# from allennlp.modules.conditional_random_field.conditional_random_field_wtrans import (
#     ConditionalRandomFieldWeightTrans,
# )
# from allennlp.modules.conditional_random_field.conditional_random_field_wlannoy import (
#     ConditionalRandomFieldWeightLannoy,
# )



import collections
import math
import re


def _postprocess(inference_output, rr_output, summary_length_percentage):
    lawbriefs_summary_map = {1: "facts", 2: "facts", 3: "arguments_petitioner", 4: "arguments_respondent",
                             5: "issue", 6: "ANALYSIS", 7: 'ANALYSIS',
                             8: 'ANALYSIS', 9: 'ANALYSIS', 10: 'ANALYSIS',
                             11: 'decision'}  ##### keys are baseline rhetorical roles and values are LawBriefs roles.
    final_categories = {'facts': 'facts', 'issue': 'issue',
                        'arguments_petitioner': 'arguments',
                        'arguments_respondent': 'arguments',
                        'ANALYSIS': 'ANALYSIS',
                        'decision': 'decision'}  ### these are the predicted sumarry sections
    additional_mandatory_categories = ['issue', 'decision']

    def get_adaptive_summary_sent_percent(sent_cnt):
        ######## get the summary sentence percentage to keep in output based in input sentence cnt. The values are found by piecewise linear regression
        if sent_cnt <= 77:
            const = 40.5421
            slope = -0.2444
        elif sent_cnt <= 122:
            const = 29.5264
            slope = -0.1013
        else:
            const = 17.8994
            slope = - 0.006
        summary_sent_precent = slope * sent_cnt + const
        return summary_sent_precent

    def get_short_rhetorical_roles(sent_list, min_token_cnt_per_rr=50):
        ####### checks if all combine sentences of a rhetorical roles are short and return list of such rhetorical roles
        rr_sents = {}  ##### keys are summary section names and values are 'token_cnt'

        #### collect all the sentenes of a summary section
        for sent_dict in sent_list:
            sent_token_cnt = len(sent_dict['sent_txt'].split())
            sent_rr = lawbriefs_summary_map[sent_dict['sent_rhetorical_role']]
            if rr_sents.get(sent_rr) is None:
                rr_sents[sent_rr] = 1
            else:
                rr_sents[sent_rr] += sent_token_cnt

        ### check which sections are short
        short_rr = []
        for rr, token_cnt in rr_sents.items():
            if token_cnt < min_token_cnt_per_rr:
                short_rr.append(rr)

        return short_rr

    def get_summary_score_threshold(sent_list):
        ####### Gives threshold for selecting summary sentences. This depends on the length of judgment
        filtered_sent_list = [sent for sent in sent_list if re.search('[a-zA-Z]', sent['sent_txt'])]

        summary_sent_precent = summary_length_percentage * 100 if summary_length_percentage != 0.0 else \
            get_adaptive_summary_sent_percent(len(filtered_sent_list))
        sents_to_keep = math.ceil(summary_sent_precent * len(filtered_sent_list) / 100)

        sent_scores = [float(i['sent_score']) for i in filtered_sent_list]
        sent_scores.sort(reverse=True)

        threshold = sent_scores[sents_to_keep - 1]
        return threshold

    def select_summary_sentences(inference_output):
        ########## selects which sentences should in summary and summary section to summary output
        for file_name, sent_list_all in inference_output.items():
            summary_score_threshold = get_summary_score_threshold(
                sent_list_all)  ### sentences above this threshold are included in summary
            short_rr = get_short_rhetorical_roles(sent_list_all)

            for sent in sent_list_all:
                in_summary = False
                sent_rr = lawbriefs_summary_map[sent['sent_rhetorical_role']]
                sent_summary_section = final_categories[sent_rr]
                if re.search('[a-zA-Z]', sent['sent_txt']):
                    if sent_rr in short_rr or sent_summary_section in additional_mandatory_categories or sent[
                        'sent_score'] >= summary_score_threshold:
                        in_summary = True

                sent['in_summary'] = in_summary
                sent['summary_section'] = sent_summary_section

    def combine_rr_summary_outputs(summary_inference_output, rr_output):
        ###### Add the summary details to rr_output inplace
        summary_details = {}  #### keys are sent_id and values are {'in_summary': True,'sent_score':0.1, 'summary_section':'ANALYSIS'}
        summary_output = list(summary_inference_output.values())[0] if list(summary_inference_output.values()) else []
        for sent in summary_output:
            in_summary_flag = sent['in_summary']
            summary_details[sent['sent_id']] = {'in_summary': in_summary_flag,
                                                'sent_score': float(sent['sent_score']),
                                                'summary_section': sent['summary_section']
                                                }

        ##### add the details to rr output
        for sent in rr_output:
            if summary_details.get(sent['id']) is not None:
                sent.update(summary_details[sent['id']])
            else:
                if sent['labels'][0] == 'PREAMBLE':
                    sent['in_summary'] = True
                    sent['summary_section'] = "PREAMBLE"
                else:
                    sent['in_summary'] = False

    def create_summary_text(rr_output):
        sectionwise_summary = collections.defaultdict(str)
        for sent in rr_output:
            if sent['in_summary']:
                sectionwise_summary[sent['summary_section']] += sent['text'] + '\n'

        return dict(sectionwise_summary)

    select_summary_sentences(inference_output)
    combine_rr_summary_outputs(inference_output, rr_output)
    summary_texts = create_summary_text(rr_output)

    return summary_texts



class InputDocument:
    """Represents a document that consists of sentences and a label for each sentence"""

    def __init__(self, sentences, labels, doc_name):
        """sentences: array of sentences labels: array of labels for each sentence """
        self.sentences = sentences
        self.labels = labels
        self.doc_name = doc_name

    def get_sentence_count(self):
        return len(self.sentences)


class DocumentsDataset:
    def __init__(self, path, max_docs=-1):
        self.path = path
        self.length = None
        self.max_docs = max_docs

    # Adapter functions for Iterator
    def __iter__(self):
        return self.readfile()

    def __len__(self):
        return self.calculate_len()

    def calculate_len(self):
        """Iterates once over the corpus to set and store length"""
        if self.length is None:
            self.length = 0
            for _ in self:
                self.length += 1

        return self.length

    def readfile(self):
        """Yields InputDocuments """
        read_docs = 0
        with open(self.path, encoding="utf-8") as f:
            sentences, tags = [], []
            doc_name = ''
            for line in f:
                if self.max_docs >= 0 and read_docs >= self.max_docs:
                    return
                line = line.strip()
                if not line:
                    if len(sentences) != 0:
                        read_docs += 1
                        yield InputDocument(sentences, tags, doc_name)
                        sentences, tags = [], []
                        doc_name = ''
                elif not line.startswith("###"):
                    ls = line.split("\t")
                    if len(ls) < 2:
                        continue
                    else:
                        tag, sentence = ls[0], ls[1]
                    sentences += [sentence]
                    tags += [tag]

                elif line.startswith("###"):
                    doc_name = line.replace("###", "").strip()



import json
import os
import shutil
import warnings
from pathlib import Path

import torch
from transformers import BertTokenizer
from wasabi import msg

import opennyai.rhetorical_roles.models as models
from opennyai.rhetorical_roles.infer_data_prep import split_into_sentences_tokenize_write, write_in_hsln_format
from opennyai.utils.download import load_model_from_cache
from .eval import infer_model
from .models import BertHSLN
from .task import pubmed_task


class RhetoricalRolePredictor():

    def __init__(self, use_gpu=True, verbose=False):
        BERT_MODEL = "bert-base-uncased"
        self.config = {
            "bert_model": BERT_MODEL,
            "bert_trainable": False,
            "model": BertHSLN.__name__,
            "cacheable_tasks": [],

            "dropout": 0.5,
            "word_lstm_hs": 758,
            "att_pooling_dim_ctx": 200,
            "att_pooling_num_ctx": 15,

            "lr": 3e-05,
            "lr_epoch_decay": 0.9,
            "batch_size": 1,
            "max_seq_length": 128,
            "max_epochs": 40,
            "early_stopping": 5,

        }
        self.use_gpu = use_gpu
        self.__verbose__ = verbose
        self.initialize()

    def initialize(self):
        """ Loads the model.pt file and initialized the model object.
        Instantiates Tokenizer for preprocessor to use
        Loads labels to name mapping file for post-processing inference response
        """
        self.CACHE_DIR = os.path.join(str(Path.home()), '.opennyai')
        self.hsln_format_txt_dirpath = os.path.join(self.CACHE_DIR, 'temp_hsln/pubmed-20k', )
        os.makedirs(self.hsln_format_txt_dirpath, exist_ok=True)

        if self.use_gpu:
            if torch.cuda.is_available():
                self.device = torch.device('cuda')
                msg.info('Rhetorical Roles will use GPU!')
            else:
                self.device = torch.device('cpu')
                msg.info('Rhetorical Roles will use CPU!')
        else:
            self.device = torch.device('cpu')
            msg.info('Rhetorical Roles will use CPU!')

        # Load model
        def create_task(create_func):
            return create_func(train_batch_size=self.config["batch_size"], max_docs=-1, data_folder=self.CACHE_DIR,
                               verbose=self.__verbose__)

        task = create_task(pubmed_task)
        self.model = getattr(models, self.config["model"])(self.config, [task])

        self.model.load_state_dict(load_model_from_cache('RhetoricalRole'), strict=False)
        self.model.to(self.device)

        # Ensure to use the same tokenizer used during training
        BERT_VOCAB = "bert-base-uncased"
        self.tokenizer = BertTokenizer.from_pretrained(BERT_VOCAB, do_lower_case=True)

    def preprocess(self, data):
        """ Preprocessing input request by tokenizing
            Extend with your own preprocessing steps as needed
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Tokenize the texts and write files
            split_into_sentences_tokenize_write(data, os.path.join(self.hsln_format_txt_dirpath, 'input_to_hsln.json'),
                                                hsln_format_txt_dirpath=self.hsln_format_txt_dirpath,
                                                verbose=self.__verbose__)

            write_in_hsln_format(os.path.join(self.hsln_format_txt_dirpath, 'input_to_hsln.json'),
                                 self.hsln_format_txt_dirpath, self.tokenizer)

            task = pubmed_task(train_batch_size=self.config["batch_size"], max_docs=-1,
                               data_folder=self.hsln_format_txt_dirpath, verbose=self.__verbose__)
            return task

    def __call__(self, data):
        """ Predict the class of a text using a trained transformer model.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            task = self.preprocess(data)
            folds = task.get_folds()
            test_batches = folds[0].test
            # metrics, confusion, labels_dict, class_report = eval_model(self.model, test_batches, self.device, task)
            labels_dict = infer_model(self.model, test_batches, self.device, task, verbose=self.__verbose__)
            filename_sent_boundries = json.load(
                open(os.path.join(self.hsln_format_txt_dirpath, 'sentece_boundries.json')))

            for doc_name, predicted_labels in zip(labels_dict['doc_names'], labels_dict['docwise_y_predicted']):
                filename_sent_boundries[doc_name]['pred_labels'] = predicted_labels

            with open(os.path.join(self.hsln_format_txt_dirpath, 'input_to_hsln.json'), 'r') as f:
                input = json.load(f)
            for file in input:
                id = str(file['id'])
                pred_id = labels_dict['doc_names'].index(id)
                pred_labels = labels_dict['docwise_y_predicted']
                annotations = file['annotations']
                for i, label in enumerate(annotations):
                    import uuid
                    uid = uuid.uuid4()
                    label_id = uid.hex + '_' + str(i)
                    label['labels'] = [pred_labels[pred_id][i]]
                    label['id'] = label_id

                file['id'] = "RhetoricalRole_" + str(file['id'])

            shutil.rmtree(self.hsln_format_txt_dirpath)  ##### remove the temporary files
            return input
